/*
 *	area.cc
 */

/*	Copyright (C) 1999, 2000 Sebastian Biallas (sb@web-factory.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include "area.h"

void tarea::init()
{
	a = NULL;
};

void tareaload(htstream *f, tarea_s **p, int level, int *left)
{
	if ((*left)<=0) {
		*p = NULL;
		return;
	}
	*p = (tarea_s *) calloc(sizeof(tarea_s), 1);
	if ((level==1) || (*left==1)) {
		f->read(&(*p)->start, sizeof(dword));
		f->read(&(*p)->end, sizeof(dword));
		(*left)--;
	} else {
		tareaload(f, &(*p)->left, level / 2, left);
		f->read(&(*p)->start, sizeof(dword));
		f->read(&(*p)->end, sizeof(dword));
		(*left)--;
		tareaload(f, &(*p)->right, level / 2, left);
	}
}

int	tarea::load(htstream *f)
{
	int count;
	f->read(&count, sizeof(int));
	tareaload(f, &a, count, &count);
	return 0;
}

void tarea::done()
{
	free_recursive(a);
}

void tareaadd(tarea_s **p, dword Start, dword End)
{
	if (*p) {
		if ((Start >= (*p)->start) && (Start <= (*p)->end)) {
			if ((*p)->end < End) (*p)->end = End;
			if ((End >= (*p)->start) && (End <= (*p)->end)) {
				if ((*p)->start > Start) (*p)->start = Start;
			}
			return;
		}
		if ((End >= (*p)->start) && (End <= (*p)->end)) {
			if ((*p)->start > Start) (*p)->start = Start;
			return;
		}
		if (Start > (*p)->end) tareaadd(&(*p)->right, Start, End);
					  else  tareaadd(&(*p)->left, Start, End);
	} else {
		// new p
		tarea_s	*tmp = (tarea_s *) malloc(sizeof(tarea_s));
		*p = tmp;
		(*p)->start = Start;
		(*p)->end = End;
		(*p)->left = NULL;
		(*p)->right = NULL;
	}
}

void tarea::add(dword Start, dword End)
{
	tareaadd(&a, Start, End);
}

bool tareacontains(tarea_s *p, dword V)
{
	if (p) {
		if (V < (p->start)) return tareacontains(p->left, V);
		if (V >= (p->end)) return tareacontains(p->right, V);
		/*if ((V >= (p->start)) && (V < (p->end)))*/ return true;
	} else return false;
}

bool tarea::contains(dword V)
{
	return tareacontains(a, V);
}

void tareafindnext(tarea_s *p, dword From, dword *res)
{
	if (p) {
		if (From < p->start) {
			*res = p->start;
			if (p->left) tareafindnext(p->left, From, res);
		} else if (From >= p->end) {
			tareafindnext(p->right, From, res);
		} else *res = From;
	} else *res = 0xffffffff;
}

dword  tarea::findnext(dword From)
{
	dword res;
	tareafindnext(a, From, &res);
	return res;
}

void tareafindprev(tarea_s *p, dword From, dword *res)
{
//FIXME ??:
	if (p) {
		if (From < p->start) {
			if (p->left) tareafindprev(p->left, From, res);
		} else if (From >= p->end) {
			*res = p->start;
			tareafindprev(p->right, From, res);
		} else *res = From;
	} else *res = 0x0;
}

dword  tarea::findprev(dword From)
{
	dword res;
	tareafindprev(a, From, &res);
	return res;
}

void tarea::free_recursive(tarea_s *p)
{
	if (p) {
		free_recursive(p->left);
		free_recursive(p->right);
		free(p);
	}
}

void tareacount(tarea_s *p, int *c)
{
	if (p) {
		tareacount(p->left, c);
		(*c)++;
		tareacount(p->right, c);
	}
}

void tareastore(htstream *f, tarea_s *p)
{
	if (p) {
		tareastore(f, p->left);
		f->write(&p->start, sizeof(dword));
		f->write(&p->end, sizeof(dword));
		tareastore(f, p->right);
	}
}

void tarea::store(htstream *f)
{
	int count = 0;
	tareacount(a, &count);
	f->write(&count, sizeof(int));
	tareastore(f, a);
}

#ifdef DEBUG
void tareadump(int sp, tarea_s *A)
{
	if (A) {
		for (int i=0; i<sp; i++) printf(" ");
		printf("%08x %08x\n", A->start, A->end);
		++sp;++sp;
		tareadump(sp, A->left);
		tareadump(sp, A->right);
	}
}

void tarea::dump()
{
	tareadump(1, a);
}
#endif
