/*
 *	fs.cc
 */

/*	Copyright (C) 1999, 2000 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "global.h"
#include "fs.h"


/*
 *	fstrXXX procs and func are used for efficient string-storing
 */
struct strnode {
	strnode   	*left;
	strnode   	*right;
	char			*str;
	int			id;
};

struct fstridpair {
	int			id;
	char			*str;
};

struct fixupentry {
	int			id;
	char			**p;
	fixupentry	*next;
};

struct fstrfile {
	htstream		*f;
	strnode		*root;
};

fstrfile		fstrfiles[1];
int			fstrfilesmax=0;
int       	fstrid=0;
fstridpair     *fstrs;
fixupentry	*fixupentrylist = NULL;
fixupentry    	*fixupentryroot = NULL;

void fsinit(htstream *f)
{
	fstrfiles[0].root = NULL;
	fstrfiles[0].f = f;
}

int fstrinsert(strnode **p, char *s)
{
	if (!*p) {
		*p = (strnode *)malloc(sizeof(strnode));
		(*p)->id = fstrid;
		(*p)->str = s;
		(*p)->left = NULL;
		(*p)->right = NULL;
		return fstrid++;
	} else {
		int i = strcmp(s, (*p)->str);
		if (i>0) {
			return fstrinsert(&(*p)->right, s);
		} else if (i==0) {
			return (*p)->id;
		} else {
			return fstrinsert(&(*p)->left, s);
		}
	}
}

void fstrput(htstream *f, char *s)
{
	int	i=fstrinsert(&fstrfiles[0].root, s);
	f->write(&i, sizeof(i));
}

void inserttofixuplist(int id, char **p)
{
	if (!fixupentryroot) {
		fixupentryroot = (fixupentry *)malloc(sizeof (fixupentry));
		fixupentrylist = fixupentryroot;
		fixupentrylist->id = id;
		fixupentrylist->p = p;
		fixupentrylist->next = NULL;
	} else {
		fixupentrylist->next = (fixupentry *)malloc(sizeof (fixupentry));
		fixupentrylist = fixupentrylist->next;
		fixupentrylist->id = id;
		fixupentrylist->p = p;
		fixupentrylist->next = NULL;
	}
}

void fstrget(htstream *f, char **p)
{
	int		id;
	f->read(&id, sizeof(id));
	inserttofixuplist(id, p);
	*p = "(unresolved string)";		// for now
}

void dofstrwrite(strnode *p, htstream *f)
{
	if (p) {
		dofstrwrite(p->left, f);
		f->write(&(p->id), sizeof p->id);
		byte length = strlen(p->str);
		f->write(&length, 1);
		if (length) f->write(p->str, length);
		dofstrwrite(p->right, f);
	}
}


void fstrflush(htstream *f)
{
	f->write(&fstrid, sizeof fstrid);
	dofstrwrite(fstrfiles[0].root, f);
	fstrid = 0;
}


void fidqsort(fstridpair *fstrs, int l, int r)
{
	int i=l,j=r;
	fstridpair *x = (fstrs + (l + r) / 2);
	int id = x->id;
	fstridpair *y;
	do {
		while ((fstrs+i)->id < id) i++;
		while (id < (fstrs+j)->id) j++;
		if (i <= j) {
			y = (fstrs+i);
			(fstrs+i)->id = (fstrs+j)->id;
			(fstrs+i)->str = (fstrs+j)->str;
			(fstrs+j)->id = y->id;
			(fstrs+j)->str = y->str;
			i++;
			j--;
		}
	} while (i <= j);
	if (l < j) fidqsort(fstrs, l, j);
	if (i < r) fidqsort(fstrs, i, r);
}

void fstrload(htstream *f)
{
	int		count, i, id;
	byte 	length;
	char      *s;

	f->read(&count, sizeof(count));
	fstrs = (fstridpair *)malloc(count*sizeof(fstridpair));
	for(i=0; i<count; i++) {
		f->read(&id, sizeof id);
		f->read(&length, sizeof length);
		s = (char *)malloc(length+1);
		if (length) f->read(s, length);
		s[length] = 0;
		(fstrs+i)->id = id;
		(fstrs+i)->str = s;
	}
	fidqsort(fstrs, 0, count-1);
	/* solve references */
	//......

	fixupentry *fixup = fixupentryroot;
	while (fixup) {
		fixupentry *temp = fixup;
		fixup = fixup->next;
		free(temp);
	}

	fixupentrylist	= NULL;
	fixupentryroot	= NULL;
	free(fstrs);
}

