/*
 *	htanaly.cpp
 */

/*	Copyright (C) 1999, 2000 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htstring.h"
#include "htanaly.h"
#include "htidle.h"
#include "htiobox.h"
#include "htdialog.h"
#include "malloc.h"
#include "httree.h"
#include "tools.h"
#include "htkeyb.h"

#include <stdlib.h>

/*
 *	CLASS ht_aviewer
 */

analy_infoline::analy_infoline(bounds *b, ht_aviewer *A, char *Format):ht_winstatictext(b, 0)
{
	analy = A;
	displayformat = strdup(Format);
	s = (char *)malloc(1024);
}

analy_infoline::~analy_infoline()
{
//	free(displayformat);
//     ht_winstatictext::~ht_winstatictext();
}

char *analy_infoline::gettext()
{
	if (valid()) {
		if (fofs!=INVALID_FILE_OFS) {
			sprintf(s, displayformat, fofs,  analy->analy->getdisasmstr(addr), " ");
		} else {
			strcpy(s, "<not in file>");
		}
		return s;
	} else {
		return "<no analyser>";
	}
}

void analy_infoline::update(ADDR cursor_addr)
{
	if (valid()) {
		fofs = analy->analy->fileaddr(cursor_addr);
		addr = cursor_addr;
	} else {
		fofs = INVALID_FILE_OFS;
		addr = INVALID_ADDR;
	}
	draw();
}

bool analy_infoline::valid()
{
	return ((analy)&&(analy->analy));
}

ht_aviewer::ht_aviewer(bounds *b, char *desc, htstream *file, int caps, analyser *Analy)
: ht_uviewer(b, desc, file, caps)
{
	analy=Analy;
	idle_count=0;
	last_active=1;
	infoline=NULL;
	register_idle_object(this);
}

ht_aviewer::~ht_aviewer()
{
	unregister_idle_object(this);
}

void ht_aviewer::attach_infoline(analy_infoline *V)
{
	infoline = V;
}

void ht_aviewer::handlemsg(htmsg *msg)
{
	char str[1024];
	if (msg->msg==htm_draw) {
		if (infoline) {
			infoline->update(cursor_ofs);
		}
	}
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case 'c':
				analy->continueanalysisat(cursor_eofs);
				clearmsg(msg);
				break;
			case 'n': {
				if (analy->validaddr(cursor_ofs, scvalid)) {
					char *n = (char *)malloc(255);
					tlabel *l = analy->getaddrlabel(cursor_ofs);
					if (l) strcpy(n, l->name); else *n = 0;
					sprintf(str, "name for address %08lx", cursor_ofs);
					if (inputbox(str, "label name:", &n, 255)) {
						if (*n) {
							if (!analy->assignlabel(cursor_ofs, n, (l)?l->type:label_loc)) {
								l = analy->findlabel(n);
								errorbox("Label `%s' already exists at address %08lx!", n, l->addr->addr);
							}
						}
					}
				} else {
					errorbox("Address %08lx is invalid!", cursor_ofs);
				}
				clearmsg(msg);
				break;
			}
/*			case 't':
				bounds b;
				b.x = 5;
				b.y = 5;
				b.w = 70;
				b.h = 15;
				ht_dialog *dlg = new ht_dialog(&b, "ein dialog");
				b.x = 4;
				b.y = 2;
				b.w -= 10;
				b.h -= 4;
				ht_static_treeview *t = new ht_static_treeview(&b, "root");
				void *n = t->add_child(0, "child1");
				   t->add_child(n, "sub1");
				   t->add_child(n, "sub2");
				n = t->add_child(0, "child2");
				   t->add_child(n, "sub3");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				   t->add_child(n, "sub4");
				t->collapse_all(0);
				t->update();
				dlg->insert(t);
				execute_dialog(dlg);
				clearmsg(msg);
				break;*/
		}
	} else if (msg->msg==htm_edit_mode) {
		analy->setdisplaymode(ANALY_EDIT_BYTES, 0);
	} else if (msg->msg==htm_view_mode) {
		analy->setdisplaymode(0, ANALY_EDIT_BYTES);
	}
	ht_uviewer::handlemsg(msg);
}

void ht_aviewer::idle()
{
	last_active=analy->active;
	analy->continueanalysis();
	if (last_active && !analy->active) {
		LOG("analyser finished after %d ops.", analy->ops_parsed);
		dirtyview();
		app->sendmsg(htm_draw, 0);
	}
	if (analy->active) {
		idle_count++;
		if (idle_count%50==0) {
			dirtyview();
			app->sendmsg(htm_draw, 0);
		}
	}
}

int  ht_aviewer::jumpstr(char *str)
{
	/* overwritten, to support symbol-jump ability */
	ADDR a;
	if (!str || !str[0]) return false;
	if ((str[0]=='\'') || (!hex2int(str, &a))) {
		if (str[0]=='\'') {
			str++;
			if (str[strlen(str)-1]!='\'') {
				errorbox("Wrong goto format!\n(That is no pending ')");
				return false;
			}
			str[strlen(str)-1]=0;
		}
		tlabel *l = analy->findlabel(str);
		if (!l) {
			errorbox("Label `%s' not found.", str);
			return false;
		}
		a = l->addr->addr;
	}
	if (!jump(a, 1, this)) {
		errorbox("%08lx is not a valid address!", a);
		return false;
	}
	return true;
}

/*
 *	CLASS hts_analy
 */

hts_analy::hts_analy(htstream *file, analyser *A, ADDR Lowestaddress, ADDR Highestaddress)
:hts_sub(file)
{
	analy = A;
	lowestaddress=Lowestaddress;
	highestaddress=Highestaddress;
}

hts_analy::~hts_analy()
{
}

void hts_analy::end_line_address(dword *ofs, int *yofs)
{
	*ofs=highestaddress;
	*yofs=0;
}

int hts_analy::getline(char *line, dword ofs, int yofs)
{
	int len;
	analy->getformattedline(ofs, yofs+1, line, &len);
	return 1;
}

int	hts_analy::next_line_address(ADDR *ofs, int *yofs, int n)
{
	int len;
	int res=0;
	while (n--) {
		if (analy->getformattedline_length(*ofs, *yofs+2, &len)) {
			(*yofs)++;
		} else {
			analy->getformattedline_length(*ofs, *yofs+1, &len);
//     		if (!analy->validaddr(*ofs+len, scvalid)) return res;
			*yofs = 0;
			*ofs += len;
		}
		res++;
	}
	return res;
}

int	hts_analy::prev_line_address(ADDR *addr, int *line, int n)
{
	int len;

	ADDR addrbuf[1024];	/* FIXME: possible buffer overflow ! */
	int	linebuf[1024];	/* FIXME: possible buffer overflow ! */
	int	i=0;
	ADDR	l=n*16;
	if (l<64) l=64;

	/* look backwards */
//     printf("============= up ==============================\n");
//     printf("%d x from %lx line %d\n", n, *addr, *line);
	ADDR tmpaddr = (*addr) - l;
	ADDR tmpaddr2;
	int	tmpline;
	start_line_address(&tmpaddr2, &tmpline);
	if ((l > (*addr)) || (tmpaddr < tmpaddr2)) tmpaddr = tmpaddr2;
	ADDR	nextaddr = tmpaddr;
	while (1) {
		if (tmpaddr >= (*addr)) {
			if ((tmpline >= (*line)) || (tmpaddr > *addr)) break;
		}
		if (analy->getformattedline_length(tmpaddr, tmpline+1, &len)) {
			nextaddr += len;
			tmpline++;
		} else {
			tmpaddr = nextaddr;
			tmpline = 0;
			continue;
		}
//          printf("%2ld. %lx line %ld\n", i, tmpaddr, tmpline-1);
		addrbuf[i] = tmpaddr;
		linebuf[i] = tmpline-1;
		i++;
	}

	if (!i) return 0;

	if (i >= n) {
//     	printf("result1: %lx line %d (%d)\n", addrbuf[i-n], linebuf[i-n], n);
		(*addr) = addrbuf[i-n];
		(*line) = linebuf[i-n];
		return n;
	} else {
//     	printf("result2: %lx line %d\n (%d)\n", addrbuf[0], linebuf[0], i);
		(*addr) = addrbuf[0];
		(*line) = linebuf[0];
		return i;
	}
}

void	hts_analy::start_line_address(dword *ofs, int *yofs)
{
	*ofs=lowestaddress;
	*yofs=0;
}

int hts_analy::jump(ADDR tofs, ADDR *ofs, int *yofs)
{
	if (analy->validaddr(tofs, scvalid)) {
		*ofs=tofs;
		*yofs=0;
		return true;
	}
	return false;
}


