/* 
 *	HT Editor
 *	htdata.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htdata.h"
#include "htdebug.h"
#include "htstring.h"
#include "tools.h"

#include <stdlib.h>
#include <string.h>

/*
 *	CLASS ht_data
 */

ht_data::ht_data()
{
}

ht_data::~ht_data()
{
}

/*
 *	CLASS ht_data_dword
 */

ht_data_dword::ht_data_dword(dword d)
{
	data=d;
}

ht_data_dword::~ht_data_dword()
{
}

/*
 *	CLASS ht_data_ptr
 */

ht_data_ptr::ht_data_ptr(void *d)
{
	data=d;
}

ht_data_ptr::~ht_data_ptr()
{
}

/*
 *	CLASS ht_data_page
 */

ht_data_page::ht_data_page()
{
	start=64;
	end=0;
}

ht_data_page::~ht_data_page()
{
}

void ht_data_page::set(int ofs, unsigned char c)
{
	if ((ofs<0) || (ofs>63)) {
		ERROR("shit2!");
	}
	start=MIN(ofs, start);
	end=MAX(ofs, end);
	buf[ofs]=c;
}

void ht_data_page::seto(int ofs, unsigned char c)
{
	if ((ofs<0) || (ofs>63)) {
		ERROR("shit2!");
	}
	start=MIN(ofs, start);
	end=MAX(ofs+1, end);
	buf[ofs]=c;
	obuf[ofs]=c;
}

unsigned char ht_data_page::get(int ofs)
{
	if ((ofs<start) || (ofs>end)) {
		ERROR("shit!");
	}
	return buf[ofs];
}

int ht_data_page::getbuf(void *buf, int ofs, int size)
{
	if (ofs<start) {
		ERROR("shit!");
	}
	int s=size;
	if (ofs+size-1>end) s-=ofs+size-1-end;
	memmove(buf, &this->buf[ofs], s);
	return s;
/*	if ((ofs<start) || (ofs+size-1>end)) {
		ERROR("shit!");
	}
	memmove(buf, &this->buf[ofs], size);*/
}

int ht_data_page::dirty(int ofs)
{
	return (obuf[ofs]!=buf[ofs]);
}

/*
 *	CLASS ht_tree
 */

ht_tree::ht_tree()
{
	root=0;
	count=0;
	cached_key=-1;
	cached_data=0;
}

ht_tree::~ht_tree()
{
	free_r(root);
}

void ht_tree::free_r(ht_node *node)
{
	if (node) {
		free_r(node->left);
		free_r(node->right);
		delete node->data;
		delete node;
	}
}

ht_node *ht_tree::get_r(ht_node *node, int key)
{
	if (!node) return 0;
	if (key>node->key) return get_r(node->right, key); else
	if (key<node->key) return get_r(node->left, key); else return node;
}

void ht_tree::set_r(ht_node **node, int key, ht_data *data)
{
	if (*node) {
		if (key>(*node)->key) set_r(&(*node)->right, key, data); else
		if (key<(*node)->key) set_r(&(*node)->left, key, data); else {
			delete (*node)->data;
			(*node)->key=key;
			(*node)->data=data;
		}
	} else {
		*node=new ht_node;
		(*node)->key=key;
		(*node)->data=data;
		(*node)->left=0;
		(*node)->right=0;
		count++;
	}
}

ht_node *ht_tree::get_smallest(ht_node *node)
{
	ht_node *s=node;
	while (s) {
		if (!s->left) break;
		s=s->left;
	}
	return s;
}

ht_node *ht_tree::get_next_r(ht_node *node, int key)
{
	ht_node *r=0;
	assert(node);
	if (key>node->key) {
		r=get_next_r(node->right, key);
		return r;
	} else if (key<node->key) {
		r=get_next_r(node->left, key);
		if (r) return r; else return node;
	}
	return get_smallest(node->right);
}

ht_node *ht_tree::enum_first()
{
	ht_node *e=get_smallest(root);
	if (e) it_key=e->key;
	return e;
}

ht_node *ht_tree::enum_next()
{
	ht_node *e=get_next_r(root, it_key);
	if (e) it_key=e->key;
	return e;
}

void ht_tree::set(int key, ht_data *data)
{
	if (cached_key==key) {
		cached_data=data;
	}
	set_r(&root, key, data);
}

ht_data *ht_tree::get(int key)
{
	if (key==cached_key) return cached_data;
	ht_node *g=get_r(root, key);
	if (g) {
		cached_key=g->key;
		cached_data=g->data;
		return g->data;
	} return 0;
}

void ht_tree::del(int key)
{
	if (cached_key==key) {
		cached_key=-1;
		cached_data=0;
	}
	ht_node *d=get_r(root, key);
	ht_node *l=d->left, *r=d->right;
	delete d->data;
	if (r) {
		memmove(d, r, sizeof(*d));
		delete r;
		get_smallest(d)->left=l;
	} else if (l) {
		memmove(d, l, sizeof(*d));
		delete l;
	} else {
		delete d;
		root=0;
	}
	count--;
}

int ht_tree::getcount()
{
	return count;
}

void ht_tree::balance()
{
	ht_node **linear_table=(ht_node**)malloc(sizeof *linear_table * count);
	ht_node **l=linear_table;
	ht_node *n=enum_first();
	int c=count;
	while (n) {
		*(l++)=n;
		c--;
		n=enum_next();
	}
	assert(!c);
	ht_node *new_root=balance_r(linear_table, 0, count);
	free_skeleton_r(root);
	delete linear_table;
	root=new_root;
/***/

/*     n=enum_first();
	while (n) {
		printf("\n--- o=%d, c=%d ---\n", n->key, (unsigned char)((ht_data_char*)n->data)->get());
		n=enum_next();
	}
	exit(-1);*/
}

ht_node *ht_tree::balance_r(ht_node **table, int l, int r)
{
	if (l<r) {
		ht_node *t=new ht_node;
		t->key=(*(table+(l+r)/2))->key;
		t->data=(*(table+(l+r)/2))->data;
		t->left=balance_r(table, l, (l+r)/2);
		t->right=balance_r(table, (l+r)/2+1, r);
		return t;
	} else {
		return 0;
	}
}

void ht_tree::free_skeleton_r(ht_node *n)
{
	if (n) {
		free_skeleton_r(n->left);
		free_skeleton_r(n->right);
		delete n;
	}
}

/*
 *   ht_collection
 */

#define COLL_INITIAL_SIZE 256
#define COLL_GROW_SIZE 512

ht_collection::ht_collection()
{
	table_maxsize=COLL_INITIAL_SIZE;
	table=(ht_data **)malloc(sizeof *table * table_maxsize);
	table_size=0;
}

ht_collection::~ht_collection()
{
	if (table) {
		for (int i=0; i<table_size; i++) {
			delete *(table+i);
		}
		delete table;
	}
}

void ht_collection::expand()
{
	table_maxsize+=COLL_GROW_SIZE;
	ht_data **s=(ht_data **)malloc(sizeof *table * table_maxsize);
	memmove(s, table, sizeof *table * table_size);
	delete table;
	table=s;
}

void ht_collection::insert(ht_data *data)
{
	if (table_size>=table_maxsize) expand();
	*(table+table_size++)=data;
}

ht_data *ht_collection::get(int index)
{
	return *(table+index);
}

void ht_collection::del(int index)
{
	delete *(table+index);
	memmove(table+index, table+index+1, sizeof *table * (table_size-index-1));
	table_size--;
}

int ht_collection::count()
{
	return table_size;
}

/*
 *   ht_string_collection
 */

#define STR_COLL_INITIAL_SIZE 256
#define STR_COLL_GROW_SIZE 512

ht_string_collection::ht_string_collection()
{
	str_table_maxsize=STR_COLL_INITIAL_SIZE;
	str_table=(char **)malloc(sizeof *str_table * str_table_maxsize);
	str_table_size=0;
}

ht_string_collection::~ht_string_collection()
{
	if (str_table) {
		for (int i=0; i<str_table_size; i++) {
			delete *(str_table+i);
		}
		delete str_table;
	}
}

void ht_string_collection::expand()
{
	str_table_maxsize+=STR_COLL_GROW_SIZE;
	char **s=(char **)malloc(sizeof *str_table * str_table_maxsize);
	memmove(s, str_table, sizeof *str_table * str_table_size);
	delete str_table;
	str_table=s;
}

void ht_string_collection::insert(char *str)
{
	if (str_table_size>=str_table_maxsize) expand();
	*(str_table+str_table_size++)=strdup(str);
}

char *ht_string_collection::get(int index)
{
	if (index<str_table_size) {
		return *(str_table+index);
	}
	return 0;
}

void ht_string_collection::del(int index)
{
	delete *(str_table+index);
	memmove(str_table+index, str_table+index+1, sizeof *str_table * (str_table_size-index-1));
	str_table_size--;
}

int ht_string_collection::count()
{
	return str_table_size;
}

/*
 *	qsort for byvalue-tables
 */

int qsortbyvalue_compare(byvalue *a, byvalue *b)
{
	return (a->key>b->key) ? 1 : (a->key<b->key) ? -1 : 0;
}

void qsortbyvalue(byvalue *table, int l, int r)
{
	int m=(l+r)/2;
	int origl=l;
	int origr=r;
	byvalue c;
	c=table[m];
	do {
		while (qsortbyvalue_compare(&table[l], &c)<0) l++;
		while (qsortbyvalue_compare(&table[r], &c)>0) r--;
		if (l<=r) {
			byvalue t;
			t=table[l];
			table[l]=table[r];
			table[r]=t;
			l++;
			r--;
		}
	} while(l<r);
	if (origl<r) qsortbyvalue(table, origl, r);
	if (l<origr) qsortbyvalue(table, l, origr);
}

/*
 *	qsort for byname-tables
 */

int qsortbyname_compare(byname *a, byname *b)
{
	int z=strncmp(a->key, b->key, MIN(a->keylen, b->keylen));
	if (!z) return (a->keylen>b->keylen) ? 1 : (a->keylen<b->keylen) ? -1 : 0; else
		return z;
}

void qsortbyname(byname *table, int l, int r)
{
	int m=(l+r)/2;
	int origl=l;
	int origr=r;
	byname c;
	c=table[m];
	do {
		while (qsortbyname_compare(&table[l], &c)<0) l++;
		while (qsortbyname_compare(&table[r], &c)>0) r--;
		if (l<=r) {
			byname t;
			t=table[l];
			table[l]=table[r];
			table[r]=t;
			l++;
			r--;
		}
	} while(l<r);
	if (origl<r) qsortbyname(table, origl, r);
	if (l<origr) qsortbyname(table, l, origr);
}

/*
 *   matchhash
 */

char *matchhash(int value, int_hash *hash_table)
{
	if (hash_table) {
		while (hash_table->desc) {
			if (hash_table->value==value) return hash_table->desc;
			hash_table++;
		}
	}
	return 0;
}

