/* 
 *	HT Editor
 *	htdebug.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "global.h"
#include "htdebug.h"

#include <stdarg.h>
#include <stdio.h>

#define CPU_CLOCK 500000000

void ht_assert_failed(char *file, int line, char *assertion)
{
	fprintf(stderr, "in file %s, line %d: assertion failed: %s\n", file, line, assertion);
	exit(1);
}

void ht_warn(char *file, int line, char *format,...)
{
	va_list arg;
	va_start(arg, format);
	fprintf(stderr, "in file %s, line %d: warning: ", file, line);
	vfprintf(stderr, format, arg);
	fputc('\n', stderr);
	va_end(arg);
}

void ht_error(char *file, int line, char *format,...)
{
	va_list arg;
	va_start(arg, format);
	fprintf(stderr, "in file %s, line %d: error: ", file, line);
	vfprintf(stderr, format, arg);
	fputc('\n', stderr);
	va_end(arg);
	exit(1);
}

#ifdef DJGPP

#define MAX_TIMERS 10

dword timer_start[MAX_TIMERS];
dword timer_end[MAX_TIMERS];
int handle[MAX_TIMERS];
int handle_count=0;

#endif

int getfreetimerhandle()
{
#ifdef DJGPP
	for (int i=0; i<handle_count; i++) {
		if (!handle[i]) return i;
	}
#endif
	return -1;
}

int new_timer()
{
#ifdef DJGPP
	int h=getfreetimerhandle();
	if (h==-1) {
		h=handle_count++;
	}
	handle[h]=1;
	timer_start[h]=0;
	timer_end[h]=0;
	return h;
#else
	return -1;     
#endif
}

void start_timer(int h)
{
#ifdef DJGPP
	dword s0, s1;
	asm("rdtsc" : "=a" (s0), "=d" (s1));
	s0=s0>>10;
	s1=s1<<22;
	timer_start[h]=s0|s1;
#else
#endif
}

void stop_timer(int h)
{
#ifdef DJGPP
	dword e0, e1;
	asm("rdtsc" : "=a" (e0), "=d" (e1));
	e0=e0>>10;
	e1=e1<<22;
	timer_end[h]=e0|e1;
#else
#endif
}

void delete_timer(int h)
{
#ifdef DJGPP
	handle[h]=0;
#endif
}

int get_timer_sec(int h)
{
#ifdef DJGPP
	return (timer_end[h]-timer_start[h])/(CPU_CLOCK/1024);
#else
	return 0;
#endif
}

int get_timer_msec(int h)
{
#ifdef DJGPP
	return (timer_end[h]-timer_start[h])/(CPU_CLOCK/1000/1024);
#else
	return 0;
#endif
}

int get_timer_tick(int h)
{
#ifdef DJGPP
	return 1024*(timer_end[h]-timer_start[h]);
#else
	return 0;
#endif
}

