/*
 *	HT Editor
 *	htdialog.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "htapp.h"
#include "htdialog.h"
#include "htidle.h"
#include "htkeyb.h"
#include "htstring.h"
#include "tools.h"

int iexecute_dialog(ht_dialog *dialog, int modal)
{
	ht_view *orig_focused=app->getselected();
	int oldx, oldy;
	screen->getcursor(&oldx, &oldy);
	dialog->setstate(ds_normal);
	app->insert(dialog);
	app->focus(dialog);
	app->sendmsg(htm_draw, 0);
	screen->show();
	while (dialog->getstate()==ds_normal) {
		if (htkeypressed()) {
			dialog->sendmsg(htm_keypressed, htgetkey());
			app->sendmsg(htm_draw, 0);
			screen->show();
		}
		if (!modal) do_idle();
	}
	int state=dialog->getstate();
	screen->setcursor(oldx, oldy);
	app->remove(dialog);
	app->focus(orig_focused);
	if (state!=ds_term_cancel) {
		return state;
	} else {
		return 0;
	}
}

int execute_dialog(ht_dialog *dialog)
{
    return iexecute_dialog(dialog, 0);
}

int execute_dialog_modal(ht_dialog *dialog)
{
    return iexecute_dialog(dialog, 1);
}

/*
 *	CLASS ht_dialog
 */

ht_dialog::ht_dialog(bounds *b, char *desc)
: ht_group(b, hto_selectable, desc)
{
	bounds c=size;
	c.x=0;
	c.y=0;
	frame=new ht_frame(&c, desc, -1);
	insert(frame);
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_dialog::~ht_dialog()
{
}

int ht_dialog::alone()
{
	return 1;
}

void ht_dialog::draw()
{
	clear(windialog_color_body);
	ht_group::draw();
}

int ht_dialog::getstate()
{
	return state;
}

void ht_dialog::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_button_pressed) {
		switch (msg->data_int) {
			case button_cancel:
				setstate(ds_term_cancel);
				clearmsg(msg);
				break;
			default:
				setstate(ds_term_ok);
				clearmsg(msg);
				break;
		}
	}
	ht_group::handlemsg(msg);
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Escape:
				setstate(ds_term_cancel);
				clearmsg(msg);
				break;
			case K_Return:
				sendmsg(htm_button_pressed, button_ok);
				clearmsg(msg);
				break;
		}
	}
}

void ht_dialog::setstate(int _state)
{
	state=_state;
}

/*
 *	CLASS ht_cluster
 */

ht_cluster::ht_cluster(bounds *b, ht_string_collection *_strings)
: ht_view(b, hto_selectable | hto_ownbuffer | hto_postprocess, "some cluster")
{
	strings=_strings;
	scount=strings->count();
	if (scount>32) scount=32;			/* we cant use more than 32... */
	sel=0;
	for (int i=0; i<scount; i++) {
		char *s=strings->get(i);
		s=strchr(s, '~');
		if (s) {
			shortcuts[i]=htkey2alt(*(s+1));
		} else shortcuts[i]=-1;
	}
}

ht_cluster::~ht_cluster()
{
}

/*
 *	CLASS ht_checkboxes
 */

ht_checkboxes::ht_checkboxes(bounds *b, ht_string_collection *strings)
: ht_cluster(b, strings)
{
	state=0;
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_checkboxes::~ht_checkboxes()
{
}

int ht_checkboxes::datasize()
{
	return sizeof (ht_checkboxes_data);
}

void ht_checkboxes::draw()
{
//	buf->fill(size.x, size.y, size.w, size.h, windialog_color_cluster_unfocused, ' ');
	clear(windialog_color_cluster_unfocused);
	int i=0;
	int vx=0, vy=0;
	int maxcolstrlen=0;
	while (i<scount) {
		int c=windialog_color_cluster_unfocused;
		if ((focused) && (sel==i)) c=windialog_color_cluster_focused;
		char *s=strings->get(i);
		int slen=strlen(s);
		if (slen>maxcolstrlen) maxcolstrlen=slen;
		if ((1<<i) & state) {
			print(vx, vy, c, "[x]");
		} else {
			print(vx, vy, c, "[ ]");
		}
		int k=0, oc=c;
		for (int q=0; q<size.w-4; q++) {
			if (!*(s+q)) break;
			if (*(s+q)=='~') {
				c=windialog_color_cluster_shortcut;
				continue;
			} else {
				printchar(vx+k+4, vy, c, *(s+q));
				k++;
			}
			c=oc;
		}
		i++;
		vy++;
		if (vy>=size.h) {
			vx+=maxcolstrlen+5;
			vy=0;
			maxcolstrlen=0;
		}
	}
}

void ht_checkboxes::getdata(void *buf)
{
	((ht_checkboxes_data*)buf)->state=state;
}

void ht_checkboxes::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_postprocess) {
		if (msg->msg==htm_keypressed) {
			for (int i=0; i<scount; i++) {
				if ((shortcuts[i]!=-1) && (msg->data_int==shortcuts[i])) {
					sel=i;
					state=state^(1<<sel);
					app->focus(this);
					dirtyview();
					clearmsg(msg);
					break;
				}
			}
		}
	} else {
		if (msg->msg==htm_keypressed) {
			switch (msg->data_int) {
				case K_Left:
					sel-=size.h;
					if (sel<0) sel=0;
					dirtyview();
					clearmsg(msg);
					break;
				case K_Right:
					sel+=size.h;
					if (sel>=scount) sel=scount-1;
					dirtyview();
					clearmsg(msg);
					break;
				case K_Up:
					sel--;
					if (sel<0) sel=0;
					dirtyview();
					clearmsg(msg);
					break;
				case K_Down:
					sel++;
					if (sel>=scount) sel=scount-1;
					dirtyview();
					clearmsg(msg);
					break;
				case K_Space:
					state=state^(1<<sel);
					dirtyview();
					clearmsg(msg);
					break;
			}
		}
	}
	ht_cluster::handlemsg(msg);
}

/*
 *	CLASS ht_radioboxes
 */

ht_radioboxes::ht_radioboxes(bounds *b, ht_string_collection *strings)
: ht_cluster(b, strings)
{
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_radioboxes::~ht_radioboxes()
{
}

int ht_radioboxes::datasize()
{
	return sizeof (ht_radioboxes_data);
}

void ht_radioboxes::draw()
{
//	buf->fill(size.x, size.y, size.w, size.h, windialog_color_cluster_unfocused, ' ');
	clear(windialog_color_cluster_unfocused);
	int i=0;
	int vx=0, vy=0;
	int maxcolstrlen=0;
	while (i<scount) {
		int c=windialog_color_cluster_unfocused;
		if ((focused) && (sel==i)) c=windialog_color_cluster_focused;
		char *s=strings->get(i);
		int slen=strlen(s);
		if (slen>maxcolstrlen) maxcolstrlen=slen;
		print(vx, vy, c, "( )");
		if (i==sel) {
			printchar(vx+1, vy, c, CHAR_RADIO);
		}
		print(vx+4, vy, c, s);
		i++;
		vy++;
		if (vy>=size.h) {
			vx+=maxcolstrlen+5;
			vy=0;
			maxcolstrlen=0;
		}
	}
}

void ht_radioboxes::getdata(void *buf)
{
	((ht_radioboxes_data*)buf)->sel=sel;
}

void ht_radioboxes::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
				sel-=size.h;
				if (sel<0) sel=0;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Right:
				sel+=size.h;
				if (sel>=scount) sel=scount-1;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Up:
				sel--;
				if (sel<0) sel=0;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down:
				sel++;
				if (sel>=scount) sel=scount-1;
				dirtyview();
				clearmsg(msg);
				break;
/*			case K_Space:
				state=state^(1<<sel);
				dirtyview();
				clearmsg(msg);
				break;*/
		}
	}
	ht_cluster::handlemsg(msg);
}

/*
 *	CLASS ht_inputfield
 */

ht_inputfield::ht_inputfield(bounds *b, int _maxtextlen)
: ht_view(b, hto_selectable, "some inputfield")
{
	textv=(byte*)malloc(_maxtextlen);
	curcharv=textv;
	textlenv=0;
	maxtextlenv=_maxtextlen;

	text=&textv;
	curchar=&curcharv;
	textlen=&textlenv;
	maxtextlen=&maxtextlenv;

	insert=1;
	ofs=0;
	attachedto=0;
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_inputfield::~ht_inputfield()
{
	freebuf();
}

void ht_inputfield::attach(ht_inputfield *inputfield)
{
	freebuf();
	inputfield->query(&curchar, &text, &textlen, &maxtextlen);
	attachedto=inputfield;
	ofs=0;
	insert=1;
}

int ht_inputfield::datasize()
{
	if (attachedto) {
		return sizeof (ht_inputfield_data);
	} else {
		return sizeof (ht_inputfield_data)+*textlen;
	}
}

void ht_inputfield::freebuf()
{
	if ((text) && (*text) && (!attachedto)) delete *text;
}

void ht_inputfield::getdata(void *buf)
{
	if (attachedto) {
		((ht_inputfield_data*)buf)->isattached=1;
		((ht_inputfield_data*)buf)->attached.attachedto=attachedto;
	} else {
		((ht_inputfield_data*)buf)->isattached=0;
		((ht_inputfield_data*)buf)->pure.textlen=*textlen;
		((ht_inputfield_data*)buf)->pure.maxtextlen=*maxtextlen;
		((ht_inputfield_data*)buf)->pure.text=(byte*)&((ht_inputfield_data*)buf)->pure.textdata;
		memmove(&((ht_inputfield_data*)buf)->pure.textdata, *text, *textlen);
	}
}

int ht_inputfield::insertbyte(byte *pos, byte b)
{
	if (*textlen<*maxtextlen) {
		memmove(pos+1, pos, *textlen-(pos-*text));
		*pos=b;
		(*textlen)++;
		dirtyview();
		return 1;
	}
	return 0;
}

void ht_inputfield::query(byte ***c, byte ***t, int **tl, int **mtl)
{
	*c=curchar;
	*t=text;
	*tl=textlen;
	*mtl=maxtextlen;
}

void ht_inputfield::setdata(void *buf)
{
	if (((ht_inputfield_data*)buf)->isattached) {
		attach(((ht_inputfield_data*)buf)->attached.attachedto);
	} else {
		textlenv=((ht_inputfield_data*)buf)->pure.textlen;
		textlen=&textlenv;
		maxtextlenv=((ht_inputfield_data*)buf)->pure.maxtextlen;
		maxtextlen=&maxtextlenv;

		freebuf();
		textv=(byte*)malloc(*maxtextlen);
		memmove(textv, ((ht_inputfield_data*)buf)->pure.text, *maxtextlen);
		text=&textv;
		curcharv=*text;
		curchar=&curcharv;
		attachedto=0;
	}
}

/*
 *	CLASS ht_strinputfield
 */

ht_strinputfield::ht_strinputfield(bounds *b, int maxstrlen)
: ht_inputfield(b, maxstrlen)
{
}

ht_strinputfield::~ht_strinputfield()
{
}

void ht_strinputfield::correct_viewpoint()
{
	if (*curchar-*text<ofs) ofs=*curchar-*text; else
	if (*curchar-*text-(size.w-2)*size.h>ofs) ofs=*curchar-*text-(size.w-2)*size.h;
}

void ht_strinputfield::draw()
{
	int c=focused ? windialog_color_input_focused : windialog_color_input_unfocused;
	byte *t=*text+ofs;
	int l=*textlen-ofs;
	if (l>size.w) l=size.w;
	int y=0;
	buf->fill(size.x, size.y, size.w, size.h, c, ' ');
	if (ofs) print(0, y, windialog_color_input_clip, "<");
	for (int k=0; k<*textlen-ofs; k++) {
		if (1+k-y*(size.w-2)>size.w-2) {
			if (y+1<size.h) y++; else break;
		}
		printchar(1+k-y*(size.w-2), y, c, *(t++));
	}
	if (*textlen-ofs > (size.w-2)*size.h) {
		print(size.w-1, y, windialog_color_input_clip, ">");
	}
	if (focused) {
		if (*curchar-*text-ofs>=(size.w-2)*size.h) setcursor(size.w-1, size.h-1); else setcursor((*curchar-*text-ofs)%(size.w-2)+1, (*curchar-*text-ofs)/(size.w-2));
	}
}

void ht_strinputfield::handlemsg(htmsg *msg)
{
	if ((msg->type==htmt_nothing) && (msg->msg==htm_keypressed)) {
		switch (msg->data_int) {
			case K_Up:
				if (*curchar-*text-ofs<size.w-2) {
					ofs-=size.w-2;
					if (ofs<0) ofs=0;
				}
				*curchar-=size.w-2;
				if (*curchar<*text) *curchar=*text;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down:
				if (*curchar-*text-ofs>=(size.w-2)*(size.h-1)) {
					ofs+=size.w-2;
				}
				*curchar+=size.w-2;
				if (*curchar-*text>*textlen) *curchar=*text+*textlen;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Left:
				if (*curchar>*text) (*curchar)--;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Right:
				if (*curchar-*text<*textlen) (*curchar)++;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_BackSpace:
				if (*curchar>*text) {
					memmove(*curchar-1, *curchar, *textlen-(*curchar-*text));
					(*textlen)--;
					(*curchar)--;
					correct_viewpoint();
					dirtyview();
					clearmsg(msg);
				}
				break;
			case K_Delete:
				if ((*curchar-*text<*textlen) && (*textlen)) {
					memmove(*curchar, *curchar+1, *textlen-(*curchar-*text)-1);
					(*textlen)--;
					dirtyview();
					clearmsg(msg);
				}
				break;
			case K_Home:
				*curchar=*text;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_End:
				*curchar=*text+*textlen;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Insert:
				insert=!insert;
				dirtyview();
				clearmsg(msg);
				break;
			default:
				if ((msg->data_int>=' ') && (msg->data_int<256)) {
					setbyte(msg->data_int);
					dirtyview();
					clearmsg(msg);
				}
		}
	}
	ht_inputfield::handlemsg(msg);
}

void ht_strinputfield::receivefocus()
{
	correct_viewpoint();
	ht_inputfield::receivefocus();
}

void ht_strinputfield::setbyte(byte a)
{
	if ((insert) || (*curchar-*text>=*textlen)) {
		if ((insertbyte(*curchar, a)) && (*curchar-*text<*textlen)) (*curchar)++;
	} else {
		**curchar=a;
		(*curchar)++;
	}
	correct_viewpoint();
}

/*
 *	CLASS ht_hexinputfield
 */

ht_hexinputfield::ht_hexinputfield(bounds *b, int maxstrlen)
: ht_inputfield(b, maxstrlen)
{
	nib=0;
	insert=1;
}

ht_hexinputfield::~ht_hexinputfield()
{
}

void ht_hexinputfield::correct_viewpoint()
{
	if (*curchar-*text<ofs) ofs=*curchar-*text; else
	if ((*curchar-*text)*3-(size.w-2)*size.h+5>ofs*3) ofs=((*curchar-*text)*3-(size.w-2)*size.h+5)/3;
}

void ht_hexinputfield::draw()
{
	int c=focused ? windialog_color_input_focused : windialog_color_input_unfocused;
	char hbuf[256], *h=hbuf;
	int y=0;
	buf->fill(size.x, size.y, size.w, size.h, c, ' ');
	if (ofs) print(0, y, windialog_color_input_clip, "<");
	int vv=*textlen-ofs;
	if (vv<0) vv=0; else if (vv>(size.w-2)*size.h/3) vv=(size.w-2)*size.h/3+1;
	for (int k=0; k<vv; k++) {
		h+=sprintf(h, "%02x ", *(*text+k+ofs));
	}
	if (vv) {
		h=hbuf;
		while ((*h) && (y<size.h)) {
			h+=lprint(1, y, c, size.w-2, h);
			y++;
		}
		y--;
	}
	if ((*textlen-ofs)*3 > (size.w-2)*size.h) {
		print(size.w-1, y, windialog_color_input_clip, ">");
	}
	if (focused) {
		if ((*curchar-*text-ofs)*3+nib+1>=(size.w-2)*size.h) setcursor(size.w-1, size.h-1); else
		setcursor(((*curchar-*text-ofs)*3+nib+1)%(size.w-2), ((*curchar-*text-ofs)*3+nib+1)/(size.w-2));
	}
}

void ht_hexinputfield::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Up:
				if (*curchar-*text-ofs<(size.w-2)/3) {
					ofs-=(size.w-2)/3;
					if (ofs<0) ofs=0;
				}
				*curchar-=(size.w-2)/3;
				if (*curchar<*text) *curchar=*text;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down:
				if (*curchar-*text-ofs>=(size.w-2)*(size.h-1)/3) {
					ofs+=(size.w-2)/3;
				}
				*curchar+=(size.w-2)/3;
				if (*curchar-*text>*textlen) *curchar=*text+*textlen;
				if (*curchar-*text>=*textlen) nib=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Left:
				if (nib) {
					nib=0;
				} else if (*curchar>*text) {
					(*curchar)--;
					nib=1;
				}
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Right:
				if (!nib) {
					if (*curchar-*text<*textlen) nib=1;
				} else if (*curchar-*text<*textlen) {
					(*curchar)++;
					nib=0;
				}
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_BackSpace:
				if (*textlen) {
					if ((nib) || (*curchar==*text)) {
						if (*curchar-*text<*textlen) {
							memmove(*curchar, *curchar+1, *textlen-(*curchar-*text)-1);
						}
					} else {
						memmove(*curchar-1, *curchar, *textlen-(*curchar-*text));
					}
					(*textlen)--;
					if (*curchar-*text) (*curchar)--;
					nib=0;
					if (*curchar-*text<*textlen) nib=1;
					correct_viewpoint();
					dirtyview();
					clearmsg(msg);
				}
				break;
			case K_Delete:
				if ((*curchar-*text<*textlen) && (*textlen)) {
					memmove(*curchar, *curchar+1, *textlen-(*curchar-*text)-1);
					(*textlen)--;
					nib=0;
					dirtyview();
					clearmsg(msg);
				}
				break;
			case K_Home:
				*curchar=*text;
				nib=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_End:
				*curchar=*text+*textlen;
				nib=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				break;
			case K_Insert:
				insert=!insert;
				dirtyview();
				clearmsg(msg);
				break;
			default:
				if ((msg->data_int>='0') && (msg->data_int<='9')) {
					setnibble(msg->data_int-'0');
					dirtyview();
					clearmsg(msg);
				} else if ((msg->data_int>='a') && (msg->data_int<='f')) {
					setnibble(msg->data_int-'a'+10);
					dirtyview();
					clearmsg(msg);
				}
		}
	}
	ht_inputfield::handlemsg(msg);
}

void ht_hexinputfield::receivefocus()
{
	correct_viewpoint();
	if ((nib) && (*curchar-*text==*textlen)) {
		nib=0;
	}
	ht_inputfield::receivefocus();
}

void ht_hexinputfield::setnibble(byte a)
{
	if (((insert) || (*curchar-*text>=*textlen)) && (nib==0)) {
		if ((insertbyte(*curchar, a<<4)) && (*curchar-*text<*textlen)) nib=1;
	} else {
		if (nib) {
			**curchar=(**curchar & 0xf0) | a;
			(*curchar)++;
			nib=0;
		} else {
			**curchar=(**curchar & 0xf) | (a<<4);
			nib=1;
		}
	}
	correct_viewpoint();
}

/*
 *	CLASS ht_button
 */

ht_button::ht_button(bounds *b, char *_text, int _value)
: ht_view(b, hto_selectable | hto_ownbuffer | hto_postprocess, "some button")
{
	value=_value;
	text=strdup(_text);
	pressed=0;
	magicchar=strchr(text, '~');
	if (magicchar) {
		int l=strlen(text);
		memmove(magicchar, magicchar+1, l-(magicchar-text));
		shortcut=htkey2alt(*magicchar);
	} else shortcut=-1;
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_button::~ht_button()
{
	if (text) delete text;
}

int ht_button::datasize()
{
	return sizeof (ht_button_data);
}

void ht_button::draw()
{
	int c=focused ? windialog_color_button_focused : windialog_color_button_unfocused;
	buf->fill(size.x, size.y, size.w-1, size.h-1, c, ' ');
	int xp=(size.w-strlen(text))/2, yp=(size.h-1)/2;
	print(xp, yp, c, text);
	if (magicchar) printchar(xp+(magicchar-text), yp, windialog_color_button_shortcut, *magicchar);
/* shadow */
	printchar(0, 1, windialog_color_button_shadow, ' ');
	for (int i=1; i<size.w-1; i++) {
		printchar(i, 1, windialog_color_button_shadow, CHAR_FILLED_HU);
	}
	printchar(size.w-1, 0, windialog_color_button_shadow, CHAR_FILLED_HL);
	printchar(size.w-1, 1, windialog_color_button_shadow, CHAR_FILLED_HU);
}

void ht_button::getdata(void *buf)
{
	((ht_button_data*)buf)->pressed=pressed;
}

void ht_button::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_postprocess) {
		if (msg->msg==htm_keypressed) {
			if (msg->data_int==shortcut) {
				push();
				dirtyview();
				clearmsg(msg);
			}
		}
	} else if (msg->type==htmt_nothing) {
		if (msg->msg==htm_keypressed) {
			switch (msg->data_int) {
				case K_Return:
				case K_Space:
					push();
					dirtyview();
					clearmsg(msg);
					break;
			}
		}
		ht_view::handlemsg(msg);
	}
}

void ht_button::push()
{
/* FIXME: wont work for encapsulated buttons... */
	group->sendmsg(htm_button_pressed, value);
	pressed=1;
}

/*
 *	CLASS ht_listbox
 */

ht_listbox::ht_listbox(bounds *b)
: ht_group(b, hto_selectable, "some listbox")
{
	bounds c;

	body=0;

	setpalette(pal_grp->get(dialog_palette_str));

	c=*b;
	c.x=c.w-1;
	c.y=0;
	c.w=1;
	scrollbar=new ht_scrollbar(&c, &windialog_color_scrollbar);
	scrollbar->disable();
	insert(scrollbar);
	
	c=*b;
	c.x=0;
	c.y=0;
	c.w--;
	body=new ht_listbox_body(&c, scrollbar);
	insert(body);
}

ht_listbox::~ht_listbox()
{
}

int ht_listbox::datasize()
{
	return body->datasize();
}

void ht_listbox::getdata(void *buf)
{
	body->getdata(buf);
}

void ht_listbox::insertstring(char *str)
{
	body->insertstring(str);
}

int ht_listbox::seek(int index)
{
	return body->seek(index);
}

void ht_listbox::setdata(void *buf)
{
	body->setdata(buf);
}

void ht_listbox::setpalette(palette *palette)
{
	if (body) body->setpalette(palette);
	ht_group::setpalette(palette);
}

/*
 *	CLASS ht_listbox_popup
 */

ht_listbox_body::ht_listbox_body(bounds *b, ht_scrollbar *_scrollbar)
: ht_view(b, hto_selectable | hto_ownbuffer, "some listbox_body")
{
	scrollbar=_scrollbar;
	list=new ht_string_collection();
	pos=0;
	cursor=0;
	setpalette(pal_grp->get(dialog_palette_str));
	vis_h_size=0;
	bounds c=*b;
	c.x+=c.w-1;
	c.w=1;
}

ht_listbox_body::~ht_listbox_body()
{
	delete list;
}

int ht_listbox_body::datasize()
{
	return sizeof (ht_listbox_data);
}

void ht_listbox_body::draw()
{
	int l=list->count();
	int d=l-pos;
	if (d>size.h) d=size.h;
	int fc=focused ? windialog_color_list_focused_unselected : windialog_color_list_unfocused_unselected;
	buf->fill(size.x, size.y, size.w, size.h, fc, ' ');
	for (int i=0; i<d; i++) {
		int c=(i==cursor) ? (focused ? windialog_color_list_focused_selected : windialog_color_list_unfocused_selected) : fc;
		if (i==cursor) {
			buf->fill(size.x+0, size.y+i, size.w, 1, c, ' ');
		}
		lprint(0, i, c, size.w, list->get(i+pos));
	}
}

void ht_listbox_body::getdata(void *buf)
{
	((ht_listbox_data*)buf)->list=list;
	((ht_listbox_data*)buf)->index=pos+cursor;
}

void ht_listbox_body::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Control_PageUp:
			case K_Home:
				pos=0;
				cursor=0;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Control_PageDown:
			case K_End: {
				pos=0;
				int c=list->count();
				cursor=c-1;
				if (cursor>=vis_h_size) {
					pos+=cursor-vis_h_size+1;
					cursor=vis_h_size-1;
				}
				dirtyview();
				clearmsg(msg);
				break;
			}
			case K_PageUp:
				cursor-=vis_h_size;
				if (cursor<0) {
					pos+=cursor;
					cursor=0;
					if (pos<0) pos=0;
				}
				dirtyview();
				clearmsg(msg);
				break;
			case K_PageDown: {
				int c=list->count();
				cursor+=vis_h_size;
				if (cursor>=vis_h_size) {
					pos+=cursor-vis_h_size+1;
					cursor=vis_h_size-1;
					if (pos+cursor>=c) pos=c-cursor-1;
				}
				dirtyview();
				clearmsg(msg);
				break;
			}
			case K_Up:
				if (cursor) {
					cursor--;
				} else {
					if (pos) pos--;
				}
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down:
				if (cursor<vis_h_size-1) {
					if (cursor<list->count()-1) cursor++;
				} else {
					if (pos+vis_h_size<list->count()) pos++;
				}
				dirtyview();
				clearmsg(msg);
				break;
		}
		int c=list->count();
		if (c>1) scrollbar->setpos((pos+cursor)*100/(c-1));
	}
	ht_view::handlemsg(msg);
}

void ht_listbox_body::insertstring(char *str)
{
	list->insert(str);
	int c=list->count();
	vis_h_size=MIN(size.h, c);
	if ((c>size.h) && (!(scrollbar->enabled))) scrollbar->enable();
}

int ht_listbox_body::seek(int index)
{
	int c=list->count();
	if (index<c) {
		cursor=index;
		pos=0;
		if (cursor>vis_h_size-1) {
			pos=cursor+1-vis_h_size;
			cursor=vis_h_size-1;
		}
		if (c>1) scrollbar->setpos((pos+cursor)*100/(c-1));
		return 1;
	}
	return 0;
}

void ht_listbox_body::setdata(void *buf)
{
	seek(((ht_listbox_data*)buf)->index);
}

/*
 *	CLASS ht_listbox_button
 */

ht_listbox_button::ht_listbox_button(bounds *b, int _value)
: ht_listbox(b)
{
	value=_value;
}

ht_listbox_button::~ht_listbox_button()
{
}

void ht_listbox_button::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Return:
			case K_Space:
				group->sendmsg(htm_button_pressed, value);
				dirtyview();
				clearmsg(msg);
				break;
		}
	}
	ht_listbox::handlemsg(msg);
}

/*
 *	CLASS ht_listpopup
 */

ht_listpopup::ht_listpopup(bounds *b)
: ht_group(b, hto_selectable, "some listpopup")
{
	bounds c;

	body=0;
	setpalette(pal_grp->get(dialog_palette_str));

	c=*b;
	c.x=c.w-1;
	c.y=0;
	c.w=1;
	scrollbar=new ht_scrollbar(&c, &windialog_color_scrollbar);
	scrollbar->disable();
	insert(scrollbar);

	c=*b;
	c.x=0;
	c.y=0;
	c.w--;
	c.h=1;
	body=new ht_listpopup_body(&c, scrollbar);
	insert(body);
}

ht_listpopup::~ht_listpopup()
{
}

int ht_listpopup::datasize()
{
	return body->datasize();
}

void ht_listpopup::getdata(void *buf)
{
	body->getdata(buf);
}

void ht_listpopup::insertstring(char *str)
{
	body->insertstring(str);
}

void ht_listpopup::setdata(void *buf)
{
	body->setdata(buf);
}

void ht_listpopup::setpalette(palette *palette)
{
	if (body) body->setpalette(palette);
	ht_group::setpalette(palette);
}

/*
 *	CLASS ht_listpopup_body
 */

#define LISTPOPUP_HEIGHT 5

ht_listpopup_body::ht_listpopup_body(bounds *b, ht_scrollbar *scrollbar)
: ht_listbox_body(b, scrollbar)
{
	disable_buffering();
	expanded=0;
}

ht_listpopup_body::~ht_listpopup_body()
{
}

void ht_listpopup_body::collapse()
{
	if (expanded) {
		expanded=0;
		if (scrollbar->enabled) {
			scrollbar->disable();
		}
		dirtyview();
	}
}

void ht_listpopup_body::draw()
{
	int c=windialog_color_list_unfocused_selected;
	if (focused) {
		c=windialog_color_list_focused_selected;
	}
	buf->fill(size.x, size.y, size.w, 1, c, ' ');
	lprint(0, 0, c, size.w-1, list->get(pos+cursor));
	int fc=focused ? windialog_color_list_focused_unselected : windialog_color_list_unfocused_unselected;
	printchar(size.w-1, 0, fc, CHAR_ARROW_DOWN);
	if (expanded) {
		buf->fill(size.x, size.y+1, size.w, 1, fc, CHAR_LINEH);
		buf->fill(size.x, size.y+2, size.w, vis_h_size, fc, ' ');
		for (int i=0; i<vis_h_size; i++) {
			int c=fc;
			if (i==cursor) {
				c=windialog_color_list_focused_selected;
				buf->fill(size.x, size.y+i+2, size.w, 1, c, ' ');
			}
			lprint(0, i+2, c, size.w, list->get(pos+i));
		}
	}
}


void ht_listpopup_body::expand()
{
	if (!expanded) {
		bounds b;
		getbounds(&b);
		b.h=MIN(LISTPOPUP_HEIGHT, list->count())+2;
		setbounds(&b);
		expanded=1;
		vis_h_size=MIN(size.h-2, list->count());
		dirtyview();
		if ((!(scrollbar->enabled)) && (list->count()>vis_h_size)) {
			bounds b;
			scrollbar->getbounds(&b);
			b.y=size.y+2;
			b.h=vis_h_size;
			scrollbar->setbounds(&b);
			scrollbar->enable();
		}
	}
}

void ht_listpopup_body::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
			case K_Right:
			case K_Up:
			case K_Down:
			case K_PageUp:
			case K_PageDown:
			case K_Control_PageUp:
			case K_Control_PageDown:
			case K_Home:
			case K_End:
				expand();
				break;
			case K_Return:
				if (expanded) {
					collapse();
					dirtyview();
					clearmsg(msg);
				}
				break;
		}
	}
	ht_listbox_body::handlemsg(msg);
}

void ht_listpopup_body::insertstring(char *str)
{
	list->insert(str);
	int c=list->count();
	vis_h_size=MIN(size.h, c);
}

void ht_listpopup_body::releasefocus()
{
	collapse();
	ht_listbox_body::releasefocus();
}

int ht_listpopup_body::seek(int index)
{
	int c=list->count();
	if (index<c) {
		int h=MIN(LISTPOPUP_HEIGHT, c);
		cursor=index;
		pos=0;
		if (cursor>h-1) {
			pos=cursor+1-h;
			cursor=h-1;
		}
		if (c>1) scrollbar->setpos((pos+cursor)*100/(c-1));
		return 1;
	}
	return 0;
}

/*
 *	CLASS ht_statictext
 */

#define STATICTEXT_MAX_AUTO_LINE_BREAK 10

ht_statictext::ht_statictext(bounds *b, char *_text, int center)
: ht_view(b, hto_ownbuffer, "some statictext")
{
	text=strdup(_text);
/* format string... */	
	linedesc=new ht_collection();
	ht_statictext_linedesc *d;
	char *t=text, *n;
	int l=strlen(text);
	while (*t) {
	    char *lastbreakable=0;
	    d=new ht_statictext_linedesc;
	    d->text=t;
	    for (int i=0; i<l+1; i++) {
		if (i>=size.w) {
		    if ((!lastbreakable) || (i-(lastbreakable-t)>STATICTEXT_MAX_AUTO_LINE_BREAK)) {
/* force line break... */
			d->len=size.w;
		    } else {
/* automatic separator line break */
			d->len=lastbreakable-t;
		    }
		    n=d->text+d->len;
		    break;
		}
		if (!*(t+i)) {
		    d->len=i;
		    n=d->text+d->len;
		    break;
		}
		char *isbreakable=strchr(" ,.?!-=", *(t+i));
		if (isbreakable) {
		    lastbreakable=t+i+1;
		} else if (*(t+i)=='\n') {
		    d->len=i;
		    n=d->text+d->len+1;
		    break;
		   }
	    }
	    if (center) d->xofs=(size.w-d->len)/2; else d->xofs=0;
	    linedesc->insert(d);
	    t=n;
	}
/**/
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_statictext::~ht_statictext()
{
	if (text) delete text;
}

void ht_statictext::draw()
{
	clear(windialog_color_body);
	int c=MIN(linedesc->count(), size.y);
	for (int i=0; i<c; i++) {
		ht_statictext_linedesc *l=(ht_statictext_linedesc*)linedesc->get(i);
		lprint(l->xofs, i, windialog_color_body, l->len, l->text);
	}
}

/*
 *	CLASS ht_listbox_ptr
 */

ht_listbox_ptr::ht_listbox_ptr(ht_listbox *_listbox)
{
	listbox=_listbox;
}

ht_listbox_ptr::~ht_listbox_ptr()
{
}

/*
 *	CLASS ht_progress_indicator
 */

ht_progress_indicator::ht_progress_indicator(bounds *b, char *format)
: ht_group(b, 0, "some progress indicator")
{
	bounds c=*b;
	c.x=0;
	c.y=0;
	insert(new ht_frame(&c, "progress", -1));
	getbounds(&c);
	c.x=1;
	c.y=1;
	c.w-=2;
	c.h-=2;
	body=new ht_progress_indicator_body(&c, format);
	insert(body);
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_progress_indicator::~ht_progress_indicator()
{
}

int ht_progress_indicator::datasize()
{
	return body->datasize();
}

void ht_progress_indicator::setdata(void *buf)
{
	body->setdata(buf);
}

/*
 *	CLASS ht_label
 */

ht_label::ht_label(bounds *b, char *_text, ht_view *_connected)
: ht_view(b, hto_postprocess, "some label")
{
	text=strdup(_text);
	magicchar=strchr(text, '~');
	if (magicchar) {
		int l=strlen(text);
		memmove(magicchar, magicchar+1, l-(magicchar-text));
	}
	connected=_connected;
	if (magicchar) {
		shortcut=htkey2alt(*magicchar);
	} else shortcut=-1;
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_label::~ht_label()
{
	if (text) delete text;
}

void ht_label::draw()
{
	lprint(0, 0, windialog_color_body, size.w, text);
	if (magicchar) printchar(magicchar-text, 0, windialog_color_shortcut, *magicchar);
}

void ht_label::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_postprocess) {
		if (msg->msg==htm_keypressed) {
			if ((shortcut!=-1) && (msg->data_int==shortcut)) {
				app->focus(connected);
				dirtyview();
				clearmsg(msg);
			}
		}
	} else ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_progress_indicator_body
 */

ht_progress_indicator_body::ht_progress_indicator_body(bounds *b, char *_format)
: ht_view(b, hto_ownbuffer | hto_selectable, "some progress indicator body")
{
	indicator=0;
	format=_format;
}

ht_progress_indicator_body::~ht_progress_indicator_body()
{
}

void ht_progress_indicator_body::ht_progress_indicator_body::draw()
{
	char buf[1024];		/* possible buffer overflow */
	clear(windialog_color_body);
	sprintf(buf, format, indicator);
	print((size.w-strlen(buf))/2, size.h/2, windialog_color_body, buf);
}

int ht_progress_indicator_body::datasize()
{
	return sizeof (ht_progress_indicator_data);
}

void ht_progress_indicator_body::setdata(void *buf)
{
	indicator=((ht_progress_indicator_data*)buf)->indicator;
	dirtyview();
}

/*
 *	CLASS ht_color_block
 */

ht_color_block::ht_color_block(bounds *b, int light_colors, int selected)
: ht_view(b, hto_ownbuffer | hto_selectable, "some color_block")
{
	color=selected;
	if (light_colors) colors=16; else colors=8;
	setpalette(pal_grp->get(dialog_palette_str));
}

ht_color_block::~ht_color_block()
{
}

int ht_color_block::datasize()
{
	return sizeof (ht_color_block_data);
}

void ht_color_block::draw()
{
	clear(windialog_color_body);
	for (int i=0; i<colors; i++) {
		int bgc=C_BLACK;
		if (i==color) {
			int fgc=focused ? C_LIGHT(C_WHITE) : C_WHITE;
			if (C_BG(windialog_color_body)==fgc) fgc=C_BLACK;
			printchar((i%4)*3, i/4, pal_grp->get_color(fgc, C_BG(windialog_color_body)), '>');
			printchar((i%4)*3+1, i/4, pal_grp->get_color(i, i), CHAR_FILLED_F);
			printchar((i%4)*3+2, i/4, pal_grp->get_color(i, bgc), CHAR_FILLED_M);
			printchar((i%4)*3+3, i/4, pal_grp->get_color(fgc, C_BG(windialog_color_body)), '<');
		} else {
			printchar((i%4)*3+1, i/4, pal_grp->get_color(i, i), CHAR_FILLED_F);
			printchar((i%4)*3+2, i/4, pal_grp->get_color(i, bgc), CHAR_FILLED_M);
		}
	}
}

void ht_color_block::getdata(void *buf)
{
	((ht_color_block_data*)buf)->color=color;
}

void ht_color_block::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
				if (color%4-1>=0) color--;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Right:
				if (color%4+1<=3) color++;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Up:
				if (color-4>=0) color-=4;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down:
				if (color+4<colors) color+=4;
				dirtyview();
				clearmsg(msg);
				break;
		}
	}
	ht_view::handlemsg(msg);
}

void ht_color_block::set_color(int i)
{
	color=i;
	dirtyview();
}

/*
 *	CLASS ht_color_manager
 */

ht_color_manager::ht_color_manager(bounds *b, palette_group *_pal_grp)
:ht_dialog(b, "colors")
{
	ipalette *i;

	orig_pal_grp=_pal_grp;
	pal_grp=orig_pal_grp->copy();

	bounds c;
/* group */
	c.x=2;
	c.y=2;
	c.w=25;
	c.h=5;
	group=new ht_listbox(&c);
	i=pal_grp->enum_first();
	while (i) {
		group->insertstring(i->name);
		i=pal_grp->enum_next();
	}
	insert(group);
/* group label */
	c.x=2;
	c.y=1;
	c.w=10;
	c.h=2;
	insert(new ht_label(&c, "~group", group));

/* items */
	c.x=2;
	c.y=9;
	c.w=25;
	c.h=5;
	items=new ht_xgroup(&c, hto_selectable, "items");
	c.x=0;
	c.y=0;
	group_to_items=new ht_collection();
	i=pal_grp->enum_first();
	while (i) {
		palette *p=i->pal;
		ht_listbox *itemlist=new ht_listbox(&c);
		group_to_items->insert(new ht_listbox_ptr(itemlist));
		for (int k=0; k<i->entries; k++) {
			itemlist->insertstring(p->name);
			p++;
		}
		items->insert(itemlist);
		i=pal_grp->enum_next();
	}
	insert(items);
/* item label */
	c.x=2;
	c.y=8;
	c.w=10;
	c.h=1;
	insert(new ht_label(&c, "~item", items));

/* foreground colors */
	c.x=30;
	c.y=3;
	c.w=13;
	c.h=4;
	fg_colors=new ht_color_block(&c, 1, 0);
	insert(fg_colors);
/* foreground label */
	c.x=31;
	c.y=2;
	c.w=10;
	c.h=1;
	insert(new ht_label(&c, "~foreground", fg_colors));

/* background colors */
	c.x=30;
	c.y=9;
	c.w=13;
	c.h=4;
	bg_colors=new ht_color_block(&c, 1, 0);
	insert(bg_colors);
/* background label */
	c.x=31;
	c.y=8;
	c.w=10;
	c.h=1;
	insert(new ht_label(&c, "~background", bg_colors));

/* select the current item */
	sendmsg(htm_keypressed, K_Up);	/* FIXME: this a hack ! */
}

ht_color_manager::~ht_color_manager()
{
	delete group_to_items;
	delete pal_grp;
	app->sendmsg(htm_dirtyview, 0);
}

int ht_color_manager::datasize()
{
	return sizeof (ht_color_manager_data);
}

void ht_color_manager::getdata(void *buf)
{
	((ht_color_manager_data*)buf)->pal_grp=pal_grp;
}

void ht_color_manager::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {

		ht_dialog::handlemsg(msg);

		if (group->isviewdirty() || items->getselected()->isviewdirty()) {
			ht_view *cur=current;
			ht_listbox_data data;
			group->getdata(&data);
			ht_listbox_ptr *p=(ht_listbox_ptr*)group_to_items->get(data.index);
			char *s=data.list->get(data.index);
			palette *q=pal_grp->get(s);
			if (q) {
				p->listbox->getdata(&data);
				char *name;
				char color;
				if (pal_grp->get_entry(q, data.index, &name, &color)) {
					fg_colors->set_color(C_FG(color));
					bg_colors->set_color(C_BG(color));
				}
			}
			if (group->isviewdirty()) {
				focus(p->listbox);
				focus(cur);
			}
		} else if (fg_colors->isviewdirty() || bg_colors->isviewdirty()) {
			ht_listbox_data data;
			group->getdata(&data);
			ht_listbox_ptr *p=(ht_listbox_ptr*)group_to_items->get(data.index);
			char *s=data.list->get(data.index);
			palette *q=pal_grp->get(s);
			if (q) {
				p->listbox->getdata(&data);
				char *name;
				char color;
				if (pal_grp->get_entry(q, data.index, &name, &color)) {
					ht_color_block_data d;
					fg_colors->getdata(&d);
					color=d.color;
					bg_colors->getdata(&d);
					color|=d.color<<4;
					pal_grp->set_entry(q, data.index, name, color);
				}
			}
		}
	} else {
		ht_dialog::handlemsg(msg);
	}
}

/*
 *	CLASS ht_search_dialog
 */

/* STATIC! */
int ht_search_dialog_lasttype=-1;
/**/

ht_search_dialog::ht_search_dialog(bounds *b)
: ht_dialog(b, "search dialog")
{
	modecount=0;

	bounds c;
	c.x=2;
	c.y=2;
	c.w=20;
	c.h=1;

	searchmodes=new ht_listpopup(&c);
	insert(searchmodes);

	c.x=2;
	c.y=1;
	c.w=4;
	c.h=1;
	insert(new ht_label(&c, "~mode", searchmodes));

	c.x=2;
	c.y=4;
	c.w=size.w-4;
	c.h=size.h-4;
	modes=new ht_xgroup(&c, hto_selectable, "modes");
	insert(modes);
}

ht_search_dialog::~ht_search_dialog()
{
//     ht_search_dialog_lasttype=mode[modeidx].id;
}

int ht_search_dialog::datasize()
{
	return sizeof (ht_search_dialog_data);
}

int ht_search_dialog::findmode(int id)
{
	for (int i=0; i<modecount; i++) {
		if (mode[i].id==id) {
			return i;
		}
	}
	return -1;
}

void ht_search_dialog::getdata(void *buf)
{
	((ht_search_dialog_data*)buf)->id=mode[modeidx].id;
	((ht_search_dialog_data*)buf)->view=mode[modeidx].view;
}

void ht_search_dialog::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		ht_dialog::handlemsg(msg);
		ht_listpopup_data data;
		searchmodes->getdata(&data);
		if (data.index!=modeidx) {
			modeidx=data.index;
			selectmode();
		}
	} else ht_dialog::handlemsg(msg);
}

void ht_search_dialog::insertmode(int id, char *desc, ht_view *v)
{
	if (modecount<MAX_SEARCH_DIALOG_MODES) {
		modes->insert(v);
		searchmodes->insertstring(desc);
		mode[modecount].id=id;
		mode[modecount].view=v;
		modecount++;
		focus(modes);
	}
}

void ht_search_dialog::selectmode()
{
	ht_listpopup_data d;
	d.index=modeidx;
	searchmodes->setdata(&d);
	sendmsg(htm_dirtyview, 0);
	modes->select(mode[modeidx].view);
}

/*
 *	INIT
 */

struct defpal {
	int idx;
	char *name;
	int color;
};

/* BLACK */
defpal windialog_black[windialog_palette_colors]={
{windialog_coloridx_body,		"body", COLOR(C_WHITE, C_BLACK)},
{windialog_coloridx_shortcut, 	"shortcut", COLOR(C_BLACK, C_WHITE)},
{windialog_coloridx_frame, 		"frame", COLOR(C_WHITE, C_BLACK)},
{windialog_coloridx_killer,		"killer", COLOR(C_GREEN, C_BLACK)},
{windialog_coloridx_scrollbar,	"scrollbar", COLOR(C_WHITE, C_BLACK)},

{windialog_coloridx_input_focused,		"input focused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_unfocused,	"input unfocused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_clip,		"input clipping chars (<,>)", COLOR(C_WHITE, C_BLUE)},

{windialog_coloridx_button_focused,	"button focused", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
{windialog_coloridx_button_unfocused,	"button unfocused", COLOR(C_BLACK, C_GREEN)},
{windialog_coloridx_button_shadow,		"button shadow", COLOR(C_LIGHT(C_BLACK), C_BLACK)},
{windialog_coloridx_button_shortcut,	"button shortcut", COLOR(C_BLACK, C_WHITE)},

{windialog_coloridx_list_focused_selected, "list focused & selected", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
{windialog_coloridx_list_focused_unselected, "list focused & unselected", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_list_unfocused_selected, "list unfocused & selected", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_list_unfocused_unselected, "list unfocused & unselected", COLOR(C_BLACK, C_CYAN)},

{windialog_coloridx_cluster_focused,	"cluster focused", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_cluster_unfocused,	"cluster unfocused", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_cluster_shortcut,	"cluster shortcut", COLOR(C_BLACK, C_CYAN)}
};

/* SPECIAL */
defpal windialog_special[windialog_palette_colors]={
{windialog_coloridx_body,		"body", COLOR(C_WHITE, C_BLACK)},
{windialog_coloridx_shortcut, 	"shortcut", COLOR(C_BLACK, C_WHITE)},
{windialog_coloridx_frame, 		"frame", COLOR(C_WHITE, C_BLACK)},
{windialog_coloridx_killer,		"killer", COLOR(C_GREEN, C_BLACK)},
{windialog_coloridx_scrollbar,	"scrollbar", COLOR(C_WHITE, C_BLACK)},

{windialog_coloridx_input_focused,		"input focused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_unfocused,	"input unfocused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_clip,		"input clipping chars (<,>)", COLOR(C_WHITE, C_BLUE)},

{windialog_coloridx_button_focused,	"button focused", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
{windialog_coloridx_button_unfocused,	"button unfocused", COLOR(C_BLACK, C_GREEN)},
{windialog_coloridx_button_shadow,		"button shadow", COLOR(C_LIGHT(C_BLACK), C_BLACK)},
{windialog_coloridx_button_shortcut,	"button shortcut", COLOR(C_BLACK, C_WHITE)},

{windialog_coloridx_list_focused_selected, "list focused & selected", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_list_focused_unselected, "list focused & unselected", COLOR(C_WHITE, C_BLACK)},
{windialog_coloridx_list_unfocused_selected, "list unfocused & selected", COLOR(C_LIGHT(C_WHITE), C_BLACK)},
{windialog_coloridx_list_unfocused_unselected, "list unfocused & unselected", COLOR(C_WHITE, C_BLACK)},

{windialog_coloridx_cluster_focused,	"cluster focused", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_cluster_unfocused,	"cluster unfocused", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_cluster_shortcut,	"cluster shortcut", COLOR(C_BLACK, C_CYAN)}
};

/* CYAN */
defpal windialog_cyan[windialog_palette_colors]={
{windialog_coloridx_body,		"body", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_shortcut, 	"shortcut", COLOR(C_LIGHT(C_YELLOW), C_CYAN)},
{windialog_coloridx_frame, 		"frame", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_killer,		"killer", COLOR(C_LIGHT(C_GREEN), C_CYAN)},
{windialog_coloridx_scrollbar,	"scrollbar", COLOR(C_CYAN, C_BLUE)},

{windialog_coloridx_input_focused,		"input focused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_unfocused,	"input unfocused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_clip,		"input clipping chars (<,>)", COLOR(C_WHITE, C_BLUE)},

{windialog_coloridx_button_focused,	"button focused", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
{windialog_coloridx_button_unfocused,	"button unfocused", COLOR(C_BLACK, C_GREEN)},
{windialog_coloridx_button_shadow,		"button shadow", COLOR(C_LIGHT(C_BLACK), C_BLACK)},
{windialog_coloridx_button_shortcut,	"button shortcut", COLOR(C_BLACK, C_WHITE)},

{windialog_coloridx_list_focused_selected, "list focused & selected", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
{windialog_coloridx_list_focused_unselected, "list focused & unselected", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_list_unfocused_selected, "list unfocused & selected", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_list_unfocused_unselected, "list unfocused & unselected", COLOR(C_BLACK, C_CYAN)},

{windialog_coloridx_cluster_focused,	"cluster focused", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_cluster_unfocused,	"cluster unfocused", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_cluster_shortcut,	"cluster shortcut", COLOR(C_LIGHT(C_YELLOW), C_CYAN)}
};

/* GRAY */
defpal windialog_gray[windialog_palette_colors]={
{windialog_coloridx_body,		"body", COLOR(C_BLACK, C_WHITE)},
{windialog_coloridx_shortcut, 	"shortcut", COLOR(C_LIGHT(C_YELLOW), C_WHITE)},
{windialog_coloridx_frame, 		"frame", COLOR(C_LIGHT(C_WHITE), C_WHITE)},
{windialog_coloridx_killer,		"killer", COLOR(C_LIGHT(C_GREEN), C_WHITE)},
{windialog_coloridx_scrollbar,	"scrollbar", COLOR(C_CYAN, C_BLUE)},

{windialog_coloridx_input_focused,		"input focused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_unfocused,	"input unfocused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
{windialog_coloridx_input_clip,		"input clipping chars (<,>)", COLOR(C_WHITE, C_BLUE)},

{windialog_coloridx_button_focused,	"button focused", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
{windialog_coloridx_button_unfocused,	"button unfocused", COLOR(C_BLACK, C_GREEN)},
{windialog_coloridx_button_shadow,		"button shadow", COLOR(C_BLACK, C_WHITE)},
{windialog_coloridx_button_shortcut,	"button shortcut", COLOR(C_LIGHT(C_YELLOW), C_GREEN)},

{windialog_coloridx_list_focused_selected, "list focused & selected", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
{windialog_coloridx_list_focused_unselected, "list focused & unselected", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_list_unfocused_selected, "list unfocused & selected", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_list_unfocused_unselected, "list unfocused & unselected", COLOR(C_BLACK, C_CYAN)},

{windialog_coloridx_cluster_focused,	"cluster focused", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
{windialog_coloridx_cluster_unfocused,	"cluster unfocused", COLOR(C_BLACK, C_CYAN)},
{windialog_coloridx_cluster_shortcut,	"cluster shortcut", COLOR(C_LIGHT(C_YELLOW), C_CYAN)}
};

int init_dialog()
{
/* black windialog */
	palette *dialog_black_palette=pal_grp->create(windialog_black_palette_str, windialog_palette_colors);
	for (int i=0; i<windialog_palette_colors; i++) {
		pal_grp->set_entry(dialog_black_palette, windialog_black[i].idx, windialog_black[i].name, windialog_black[i].color);
	}
/* special dialogs */
	palette *dialog_special_palette=pal_grp->create(windialog_special_palette_str, windialog_palette_colors);
	for (int i=0; i<windialog_palette_colors; i++) {
		pal_grp->set_entry(dialog_special_palette, windialog_special[i].idx, windialog_special[i].name, windialog_special[i].color);
	}
/* cyan dialogs */
	palette *dialog_cyan_palette=pal_grp->create(windialog_cyan_palette_str, windialog_palette_colors);
	for (int i=0; i<windialog_palette_colors; i++) {
		pal_grp->set_entry(dialog_cyan_palette, windialog_cyan[i].idx, windialog_cyan[i].name, windialog_cyan[i].color);
	}
/* gray dialogs */
	palette *dialog_gray_palette=pal_grp->create(windialog_gray_palette_str, windialog_palette_colors);
	for (int i=0; i<windialog_palette_colors; i++) {
		pal_grp->set_entry(dialog_gray_palette, windialog_gray[i].idx, windialog_gray[i].name, windialog_gray[i].color);
	}
	return 1;
}

/*
 *	DONE
 */

int done_dialog()
{
	return 1;
}

