/* 
 *	HT Editor
 *	htdialog.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTDIALOG_H__
#define __HTDIALOG_H__

#include "htdebug.h"
#include "htobj.h"

/*
 *	Palette entries
 */

#define window_palette_str					windialog_black_palette_str
#define dialog_palette_str					windialog_gray_palette_str

#define windialog_cyan_palette_str				"windialog (cyan)"
#define windialog_black_palette_str				"windialog (black)"
#define windialog_blue_palette_str				"windialog (blue)"
#define windialog_gray_palette_str				"windialog (gray)"
#define windialog_special_palette_str				"windialog (special)"

#define windialog_palette_colors           			19

#define windialog_coloridx_body					0
#define windialog_coloridx_shortcut				1
#define windialog_coloridx_frame				2
#define windialog_coloridx_killer				3
#define windialog_coloridx_scrollbar				4

#define windialog_coloridx_input_focused			5
#define windialog_coloridx_input_unfocused			6
#define windialog_coloridx_input_clip 				7

#define windialog_coloridx_button_focused			8
#define windialog_coloridx_button_unfocused			9
#define windialog_coloridx_button_shadow 			10
#define windialog_coloridx_button_shortcut			11

#define windialog_coloridx_list_focused_selected		12
#define windialog_coloridx_list_focused_unselected		13
#define windialog_coloridx_list_unfocused_selected		14
#define windialog_coloridx_list_unfocused_unselected		15

#define windialog_coloridx_cluster_focused			16
#define windialog_coloridx_cluster_unfocused			17
#define windialog_coloridx_cluster_shortcut			18

#define windialog_color_body						pal[windialog_coloridx_body].value
#define windialog_color_shortcut					pal[windialog_coloridx_shortcut].value
#define windialog_color_frame						pal[windialog_coloridx_frame].value
#define windialog_color_killer					pal[windialog_coloridx_killer].value
#define windialog_color_scrollbar					pal[windialog_coloridx_scrollbar].value

#define windialog_color_input_focused				pal[windialog_coloridx_input_focused].value
#define windialog_color_input_unfocused				pal[windialog_coloridx_input_unfocused].value
#define windialog_color_input_clip					pal[windialog_coloridx_input_clip].value

#define windialog_color_button_focused				pal[windialog_coloridx_button_focused].value
#define windialog_color_button_unfocused			pal[windialog_coloridx_button_unfocused].value
#define windialog_color_button_shadow				pal[windialog_coloridx_button_shadow].value
#define windialog_color_button_shortcut				pal[windialog_coloridx_button_shortcut].value

#define windialog_color_list_focused_selected			pal[windialog_coloridx_list_focused_selected].value
#define windialog_color_list_focused_unselected			pal[windialog_coloridx_list_focused_unselected].value
#define windialog_color_list_unfocused_selected			pal[windialog_coloridx_list_unfocused_selected].value
#define windialog_color_list_unfocused_unselected		pal[windialog_coloridx_list_unfocused_unselected].value

#define windialog_color_cluster_focused				pal[windialog_coloridx_cluster_focused].value
#define windialog_color_cluster_unfocused			pal[windialog_coloridx_cluster_unfocused].value
#define windialog_color_cluster_shortcut			pal[windialog_coloridx_cluster_shortcut].value

/*
 *	CLASS ht_dialog
 */

#define ds_normal			0
#define ds_term_ok			1
#define ds_term_cancel		2

#define button_cancel		0
#define button_ok			1
#define button_yes            button_ok
#define button_no			2

class ht_dialog: public ht_group {
private:
	int state;
	ht_frame *frame;
public:
	ht_dialog(bounds *b, char *desc);
	~ht_dialog();
/* overwritten */
	virtual	int alone();
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	int getstate();
	virtual	void setstate(int state);
};

/*
 *	CLASS ht_cluster
 */

class ht_cluster: public ht_view {
protected:
	ht_string_collection *strings;
	int sel;
	int scount;
	int shortcuts[32];
public:
	ht_cluster(bounds *b, ht_string_collection *strings);
	~ht_cluster();
};

/*
 *	CLASS ht_checkboxes
 */

struct ht_checkboxes_data {
	dword state;
};

class ht_checkboxes: public ht_cluster {
protected:
	dword state;
public:
	ht_checkboxes(bounds *b, ht_string_collection *strings);
	~ht_checkboxes();
/* overwritten */
	virtual	int datasize();
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void getdata(void *buf);
};

/*
 *	CLASS ht_radioboxes
 */


struct ht_radioboxes_data {
	dword sel;
};

class ht_radioboxes: public ht_cluster {
public:
	ht_radioboxes(bounds *b, ht_string_collection *strings);
	~ht_radioboxes();
/* overwritten */
	virtual	int datasize();
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void getdata(void *buf);
};

/*
 *	CLASS ht_inputfield
 */

class ht_inputfield;

struct ht_inputfield_data {
	int isattached;
	union {
		struct {
			ht_inputfield *attachedto;
		} attached;
		struct {
			int textlen;
			int maxtextlen;
			byte *text;
			byte textdata[0];
		} pure;
	};
};

class ht_inputfield: public ht_view {
protected:
	byte **text, *textv;
	byte **curchar, *curcharv;
	int *textlen, textlenv;
	int *maxtextlen, maxtextlenv;
	int insert;
	int ofs;
	ht_inputfield *attachedto;

			void freebuf();
			int insertbyte(byte *pos, byte b);
public:
	ht_inputfield(bounds *b, int maxtextlen);
	virtual ~ht_inputfield();
/* overwritten */
	virtual	int datasize();
	virtual	void getdata(void *buf);
	virtual	void setdata(void *buf);
/* new */
			void attach(ht_inputfield *inputfield);
			void query(byte ***curchar, byte ***text, int **textlen, int **maxtextlen);
};

/*
 *	CLASS ht_strinputfield
 */

#define ht_strinputfield_data ht_inputfield_data

class ht_strinputfield: public ht_inputfield {
protected:
			void correct_viewpoint();
public:
	ht_strinputfield(bounds *b, int maxstrlen);
	virtual ~ht_strinputfield();
/* overwritten */
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void receivefocus();
/* new */
			void setbyte(byte a);
};

/*
 *	CLASS ht_hexinputfield
 */

#define ht_hexinputfield_data ht_inputfield_data

class ht_hexinputfield: public ht_inputfield {
protected:
	int nib;

			void correct_viewpoint();
public:
	ht_hexinputfield(bounds *b, int maxstrlen);
	virtual ~ht_hexinputfield();
/* overwritten */
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void receivefocus();
/* new */
			void setnibble(byte a);
};

/*
 *	CLASS ht_button
 */

struct ht_button_data {
	int pressed;
};

class ht_button: public ht_view {
protected:
	int value;
	int pressed;
	char *text;
	char *magicchar;
	int shortcut;
public:
	ht_button(bounds *b, char *text, int value);
	virtual ~ht_button();
/* overwritten */
	virtual	int datasize();
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	void push();
};

/*
 *	CLASS ht_listbox_body
 */

struct ht_listbox_data {
	ht_string_collection *list;
	int index;
};

class ht_listbox_body: public ht_view {
protected:
	ht_string_collection *list;
	int pos, cursor;
	int vis_h_size;
	ht_scrollbar *scrollbar;
public:
	ht_listbox_body(bounds *b, ht_scrollbar *scrollbar);
	virtual ~ht_listbox_body();
/* overwritten */
	virtual	int datasize();
	virtual	void draw();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
	virtual	void setdata(void *buf);
/* new */
	virtual	void insertstring(char *str);
	virtual	int seek(int index);
};

/*
 *	CLASS ht_listbox
 */

class ht_listbox: public ht_group {
protected:
	ht_listbox_body *body;
	ht_scrollbar *scrollbar;
public:
	ht_listbox(bounds *b);
	~ht_listbox();
/* overwritten */
	virtual	int datasize();
	virtual	void getdata(void *buf);
	virtual	void setdata(void *buf);
	virtual	void setpalette(palette *palette);
/* new */
	void insertstring(char *str);
	int seek(int index);
};

/*
 *	CLASS ht_listbox_button
 */

class ht_listbox_button: public ht_listbox {
private:
	int value;
public:
	ht_listbox_button(bounds *b, int value);
	virtual ~ht_listbox_button();
/* overwritten */
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_listpopup_body
 */

#define ht_listpopup_data ht_listbox_data

class ht_listpopup_body: public ht_listbox_body {
protected:
	int expanded;
public:
	ht_listpopup_body(bounds *b, ht_scrollbar *scrollbar);
	virtual ~ht_listpopup_body();
/* overwritten */
			void collapse();
	virtual	void draw();
			void expand();
	virtual	void handlemsg(htmsg *msg);
	virtual	void insertstring(char *str);
	virtual	void releasefocus();
	virtual	int seek(int index);
};

/*
 *	CLASS ht_listpopup
 */

class ht_listpopup: public ht_group {
protected:
	ht_listbox_body *body;
	ht_scrollbar *scrollbar;
public:
	ht_listpopup(bounds *b);
	~ht_listpopup();
/* overwritten */
	virtual	int datasize();
	virtual	void getdata(void *buf);
	virtual	void setdata(void *buf);
	virtual	void setpalette(palette *palette);
/* new */
	void insertstring(char *str);
};

/*
 *	CLASS ht_statictext
 */
 
class ht_statictext_linedesc: public ht_data {
public:
    int xofs;
    int len;
    char *text;
}; 

class ht_statictext: public ht_view {
private:
	char *text;
	ht_collection *linedesc;
public:
	ht_statictext(bounds *b, char *text, int center);
	virtual ~ht_statictext();
/* overwritten */
	virtual 	void draw();
};

/*
 *	CLASS ht_label
 */

class ht_label: public ht_view {
protected:
	ht_view *connected;
	char *text;
	char *magicchar;
	int shortcut;
public:
	ht_label(bounds *b, char *text, ht_view *connected);
	~ht_label();
/* overwritten */
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_progress_indicator_body
 */

class ht_progress_indicator_body: public ht_view {
protected:
	char *format;
	int indicator;
public:
	ht_progress_indicator_body(bounds *b, char *format);
	~ht_progress_indicator_body();
/* overwritten */
	virtual	int datasize();
	virtual	void draw();
	virtual	void setdata(void *buf);
};

/*
 *	CLASS ht_progress_indicator
 */

struct ht_progress_indicator_data {
	int indicator;
};

class ht_progress_indicator: public ht_group {
protected:
	ht_progress_indicator_body *body;
public:
	ht_progress_indicator(bounds *b, char *format);
	virtual ~ht_progress_indicator();
/* overwritten */
	virtual	int datasize();
	virtual	void setdata(void *buf);
};

/*
 *	CLASS ht_color_block
 */

struct ht_color_block_data {
	int color;
};

class ht_color_block: public ht_view {
protected:
	int color;
	int colors;
public:
	ht_color_block(bounds *b, int light_colors, int selected);
	~ht_color_block();
/* overwritten */
	virtual	int datasize();
	virtual	void draw();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	void set_color(int i);
};

/*
 *	CLASS ht_listbox_ptr
 */

class ht_listbox_ptr: public ht_data {
public:
	ht_listbox *listbox;
	ht_listbox_ptr(ht_listbox *listbox);
	~ht_listbox_ptr();
};

/*
 *	CLASS ht_color_manager
 */

struct ht_color_manager_data {
	palette_group *pal_grp;
};

class ht_color_manager: public ht_dialog {
protected:
	ht_listbox *group;
	ht_xgroup *items;
	ht_collection *group_to_items;
	ht_color_block *fg_colors;
	ht_color_block *bg_colors;
	palette_group *orig_pal_grp;
	palette_group *pal_grp;
public:
	ht_color_manager(bounds *b, palette_group *pal);
	~ht_color_manager();
/* overwritten */
	virtual	int datasize();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_search_dialog
 */

struct ht_search_dialog_data {
	int id;
	ht_view *view;
};

struct ht_search_dialog_mode {
	int id;
	ht_view *view;
};

#define MAX_SEARCH_DIALOG_MODES 16

class ht_search_dialog: public ht_dialog {
protected:
	ht_listpopup *searchmodes;
	ht_xgroup *modes;

	int modeidx;
	ht_search_dialog_mode mode[MAX_SEARCH_DIALOG_MODES];
	int modecount;

	int findmode(int id);
	void selectmode();
public:
	ht_search_dialog(bounds *b);
	~ht_search_dialog();
/* overwritten */
	virtual	int datasize();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
/* new */
			void insertmode(int id, char *desc, ht_view *v);
};

/*
 *	misc
 */

int execute_dialog(ht_dialog *dialog);
int execute_dialog_modal(ht_dialog *dialog);

/*
 *	INIT
 */

int init_dialog();

/*
 *	DONE
 */

int done_dialog();

#endif /* !__HTDIALOG_H__ */

