/* 
 *	HT Editor
 *	htiobox.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htdialog.h"
#include "htiobox.h"
#include "tools.h"

#include <stdlib.h>
#include <string.h>

int imsgbox(bounds *b, int buttonmask, char *title, char *buf)
{
	int strl=strlen(buf);
	ht_dialog *dialog=new ht_dialog(b, title);

	bounds c;

	if (strl<b->w-4) {
		c.w=strl;
		c.h=1;
		c.x=(b->w-c.w)/2;
		c.y=1;
	} else {
		c.x=2;
		c.y=1;
		c.w=b->w-4;
		c.h=b->h-4;
	}
	ht_statictext *text=new ht_statictext(&c, buf, 1);
	dialog->insert(text);

	int buttons=0;
	if (buttonmask & msgbox_ok) buttons++;
	if (buttonmask & msgbox_cancel) buttons++;
	if (buttonmask & msgbox_yes) buttons++;
	if (buttonmask & msgbox_no) buttons++;

	int pos=0;
	if (buttonmask & msgbox_ok) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-3;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button(&c, "O~K", button_ok);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_yes) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-3;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button(&c, "~Yes", button_yes);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_no) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-3;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button(&c, "~No", button_no);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_cancel) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-3;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button(&c, "Cancel", button_cancel);
		dialog->insert(button);
		pos++;
	}

	int r=execute_dialog(dialog);
	delete dialog;
	return r;
}

int msgbox(int buttonmask, char *title, char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	int strl=strlen(buf);

	bounds b;
	app->getbounds(&b);
	b.w=45;
	b.h=strl / (b.w-4) + 6;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	return imsgbox(&b, buttonmask, title, buf);
}

int msgboxrect(bounds *b, int buttonmask, char *title, char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);

	return imsgbox(b, buttonmask, title, buf);
}
#if 0
int errorbox(char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return msgbox(msgbox_ok, "error", buf);
}

int infobox(char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return msgbox(msgbox_ok, "information", buf);
}

int warnbox(char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return msgbox(msgbox_ok, "warning", buf);
}
#endif
int inputbox(char *title, char *Label, char **result, int limit)
{
	bounds b;
	app->getbounds(&b);
	b.x = (b.w - 60) / 2,
	b.y = (b.h - 8) / 2;
	b.w = 60;
	b.h = 8;
	return inputboxrect(&b, title, Label, result, limit);
}

int inputboxrect(bounds *b, char *title, char *Label, char **result, int limit)
{
	ht_dialog *dialog=new ht_dialog(b, title);
	ht_view *view;

	bounds  b2;
	b2.x = 4 + strlen(Label);
	b2.y = 2;
	b2.w = b->w - 3 - b2.x;
	b2.h = 1;

	ht_strinputfield *input=new ht_strinputfield(&b2, limit);
	ht_inputfield_data d;
	d.isattached=0;
	d.pure.text=(byte*)*result;
	d.pure.textlen=strlen(*result);
	d.pure.maxtextlen=32;
	input->setdata(&d);
	dialog->insert(input);

	b2.x = 2;
	b2.y = 2;
	b2.w = 3 + strlen(Label) - b2.x;
	b2.h = 1;

	view=new ht_statictext(&b2, Label, 0);
	dialog->insert(view);

	b2.x = b->w - 24;
	b2.y = b->h - 4;
	b2.w = 10;
	b2.h = 2;

	view=new ht_button(&b2, "O~k", button_ok);
	dialog->insert(view);

	b2.x += 12;

	view=new ht_button(&b2, "Cancel", button_cancel);
	dialog->insert(view);

	if (execute_dialog(dialog)) {
		ht_inputfield_data *data=(ht_inputfield_data*)malloc(input->datasize());
		input->getdata(data);
		free(*result);
		*result=(char*)malloc(data->pure.textlen+1);
		memmove(*result, data->pure.text, data->pure.textlen);
		*(*result+data->pure.textlen)=0;
		delete data;
		delete dialog;
		return true;
	}
	delete dialog;
	return false;
}

void get_std_progress_indicator_metrics(bounds *b)
{
	app->getbounds(b);
	b->x=b->w/10;
	b->y=b->h*2/5;
	b->w=b->w*4/5;
	b->h=b->h/5;
}

