/* 
 *	HT Editor
 *	htpe.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpehead.h"
#include "htpeexp.h"
#include "htpeimp.h"
#include "htpedimp.h"
#include "htpeimg.h"
#include "htperes.h"
#include "htstream.h"

#include <stdlib.h>
#include <string.h>

int htpe_detect(htstream *file)
{
	dword pemagic;
	file->seek(get_newexe_header_ofs(file));
	file->read(&pemagic, 4);
	return (pemagic==PE_MAGIC);
}

ht_view *htpe_init(bounds *b, htstream *file, ht_group *group)
{
	ht_xgroup *g=new ht_pe(b, file);
	insert_int(g, file, &htpeheader_int);
	insert_int(g, file, &htpeexports_int);
	insert_int(g, file, &htpeimports_int);
	insert_int(g, file, &htpedelayimports_int);
	insert_int(g, file, &htpeimage_int);
	insert_int(g, file, &htperesources_int);
	return g;
}

viewer_int htpe_int = {
	htpe_detect,
	htpe_init
};

ht_pe::ht_pe(bounds *b, htstream *file)
: ht_xgroup(b, hto_browsable | hto_selectable, DESC_PE)
{
	int h=get_newexe_header_ofs(file);
	LOG("%s: PE: found header at %08x", file->getname(), h);
	pe_shared=new ht_pe_shared_data;
	pe_shared->is_pe32=0;
	pe_shared->hdr_ofs=h;
	pe_shared->v_image=0;
	pe_shared->v_dimports=0;
	pe_shared->v_imports=0;
	pe_shared->v_exports=0;
	pe_shared->v_header=0;

/* headers */
	file->seek(h);
	file->read(&pe_shared->coffheader, sizeof pe_shared->coffheader);
	file->read(&pe_shared->pe32header, sizeof pe_shared->pe32header);

/* read section headers */
	int os=0;

	/* 1. get section count */
	file->seek(h+6);
	pe_shared->sections.section_count=0;
	file->read(&pe_shared->sections.section_count, 2);
	/* 2. get size of optional header */
	file->seek(h+20);
	file->read(&os, 2);
	/* 3. read sections */
	file->seek(h+os+24);
	pe_shared->sections.sections=(PE_SECTION_HEADER*)malloc(pe_shared->sections.section_count * sizeof *pe_shared->sections.sections);
	file->read(pe_shared->sections.sections, pe_shared->sections.section_count*sizeof *pe_shared->sections.sections);

	shared_data=pe_shared;
}

ht_pe::~ht_pe()
{
}

/*
 *	rva conversion routines
 */

int pe_rva_to_section(pe_section_headers *section_headers, ADDR rva, int *section)
{
	PE_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((rva>=s->data_address) &&
		(rva<s->data_address+s->data_size)) {
			*section=i;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_rva_to_ofs(pe_section_headers *section_headers, ADDR rva, dword *ofs)
{
	PE_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((rva>=s->data_address) &&
		(rva<s->data_address+s->data_size)) {
			*ofs=rva-s->data_address+s->data_offset;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_rva_is_valid(pe_section_headers *section_headers, ADDR rva)
{
	PE_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((rva>=s->data_address) &&
		(rva<s->data_address+s->data_size)) {
			return 1;
		}
		s++;
	}
	return 0;
}

/*
 *	ofs conversion routines
 */

int pe_ofs_to_rva(pe_section_headers *section_headers, dword ofs, ADDR *rva)
{
	PE_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((ofs>=s->data_offset) &&
		(ofs<s->data_offset+s->data_size)) {
			*rva=ofs-s->data_offset+s->data_address;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_ofs_to_section(pe_section_headers *section_headers, dword ofs, ADDR *section)
{
	PE_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((ofs>=s->data_offset) &&
		(ofs<s->data_offset+s->data_size)) {
			*section=i;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_ofs_to_rva_and_section(pe_section_headers *section_headers, dword ofs, ADDR *rva, ADDR *section)
{
	int r=pe_ofs_to_rva(section_headers, ofs, rva);
	if (r) {
		r=pe_ofs_to_section(section_headers, ofs, section);
	}
	return r;
}

int pe_ofs_is_valid(pe_section_headers *section_headers, dword ofs)
{
	ADDR rva;
	return pe_ofs_to_rva(section_headers, ofs, &rva);
}

