/* 
 *	HT Editor
 *	htpeexp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htdata.h"
#include "htdebug.h"
#include "htiobox.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpeexp.h"
#include "htstream.h"
#include "htstring.h"
#include "httag.h"
#include "tools.h"
#include "viewers.h"

#include <stdlib.h>
#include <string.h>

#define PE_EXPORT_DESC_LINES 1	/* important ! set to the number
							   of descriptive lines in the beginning
							   of the (hts_)mask */

int htpeexports_detect(htstream *file)
{
	int h=get_newexe_header_ofs(file);
	dword pe=0;
	file->seek(h);
	file->read(&pe, 2);
	if (pe==PE_MAGIC) {
		int e, f;
		file->seek(h+0x78+PE_DIRECTORY_ENTRY_EXPORT*8);
		file->read(&e, 4);
		file->read(&f, 4);
		return e|f;
	}
	return 0;
}

ht_view *htpeexports_init(bounds *b, htstream *file, ht_group *group)
{
	int h0=new_timer();
	start_timer(h0);
	ht_group *g;

	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;
	int h=get_newexe_header_ofs(file);
	char eline[1024];   /* FIXME: possible buffer overflow */
	char *elinep;
	dword *efunct=0, *enamet=0;
	word *eordt=0;
	bounds c;
	ht_winstatictext *head;
	char *filename=file->getname();

	c=*b;
	g=new ht_group(&c, 0, "pe/exports-g");

	c.y++;
	c.h--;
	ht_uviewer *v=new ht_pe_export_viewer(&c, DESC_PE_EXPORTS, file, 0);
	hts_pe_export_mask *m=new hts_pe_export_mask(file, h);

	c.y--;
	c.h=1;

/* get export directory offset */
	/* 1. get export directory rva */
	dword eofs;
	dword erva=pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_EXPORT].address;
	dword esize=pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_EXPORT].size;
	/* 2. transform it into an offset */
	if (!pe_rva_to_ofs(&pe_shared->sections, erva, &eofs)) goto pe_read_error;
	LOG("%s: PE: reading export directory at offset %08x, rva %08x, size %08x...", filename, eofs, erva, esize);

	char info[128];
	sprintf(info, "* PE export directory at offset %08lx", eofs);
	m->add_mask(info);

/* make a memfile out of this section */
	htstream *efile;
	char *esectionbuf;
	efile=new htmemstream(eofs, esize);
	esectionbuf=(char*)malloc(esize);
	file->seek(eofs);
	file->read(esectionbuf, esize);
	efile->write(esectionbuf, esize);
	file=efile;
	delete esectionbuf;
/* read export directory header */
	PE_EXPORT_DIRECTORY edir;
	file->seek(eofs);
	file->read(&edir, sizeof edir);
/* get export filename */
	dword ename_ofs;
	char *ename;
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.name_address, &ename_ofs)) goto pe_read_error;
	file->seek(ename_ofs);
	ename=getstrz(file);
/* ...and add it to the mask */
	sprintf(eline, "Ord  VA       Name      %s exports %ld functions", ename, edir.function_count);
	head=new ht_winstatictext(&c, eline);
	delete ename;
/* read in function entrypoint table */
	dword efunct_ofs;
	efunct=(dword*)malloc(edir.function_count*sizeof *efunct);
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.function_table_address, &efunct_ofs)) goto pe_read_error;
	file->seek(efunct_ofs);
	file->read(efunct, edir.function_count*sizeof *efunct);
/* read in name address table */
	dword enamet_ofs;
	enamet=(dword*)malloc(edir.name_count*sizeof *efunct);
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.name_table_address, &enamet_ofs)) goto pe_read_error;
	file->seek(enamet_ofs);
	file->read(enamet, edir.name_count*sizeof *efunct);
/* read in ordinal table */
	dword eordt_ofs;
	eordt=(word*)malloc(edir.name_count*sizeof *efunct);
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.ordinal_table_address, &eordt_ofs)) goto pe_read_error;
	file->seek(eordt_ofs);
	file->read(eordt, edir.name_count*sizeof *efunct);

	int *lord;
	lord=(int*)malloc(sizeof *lord*edir.function_count);
	memset(lord, 0, sizeof *lord*edir.function_count);
	for (dword i=0; i<edir.name_count; i++) {
		*(lord+*(eordt+i))=1;
	}

/* exports by ordinal */
	for (dword i=0; i<edir.function_count; i++) {
		if (*(lord+i)) continue;

		dword f=*(efunct+i)+pe_shared->pe32header.image_base;

		elinep=eline;
		elinep+=sprintf(elinep, "%04lx %08lx ", i+edir.ordinal_base, f);
		mktag_ref(elinep, i, tag_coloridx_sel_unselected, "<ordinal only>", f);

		ht_pe_export_function *efd=new ht_pe_export_function(f, i+edir.ordinal_base);
		pe_shared->exports.funcs.insert(efd);
		m->add_mask(eline);
	}
	delete lord;
/* exports by name */
	for (dword i=0; i<edir.name_count; i++) {
		dword o=*(eordt+i);
		dword f=*(efunct+o)+pe_shared->pe32header.image_base;
		dword en;
		if (!pe_rva_to_ofs(&pe_shared->sections, *(enamet+i), &en)) goto pe_read_error;
		file->seek(en);
		char *s=getstrz(file);

		elinep=eline;
		elinep+=sprintf(elinep, "%04lx %08lx ", o, f);
		mktag_ref(elinep, o, tag_coloridx_sel_unselected, s, f);

		ht_pe_export_function *efd=new ht_pe_export_function(f, o, s);
		pe_shared->exports.funcs.insert(efd);
		delete s;
		m->add_mask(eline);
	}

	if (!edir.function_count) m->add_mask(_SEL("ffffffff", "17")"<no exported functions>");

	v->insertsub(m);

	g->insert(head);
	g->insert(v);

	stop_timer(h0);
	LOG("%s: PE: it took %d ticks (%d msec) to read the exports", filename, get_timer_tick(h0), get_timer_msec(h0));
	delete_timer(h0);

	if (efunct) delete efunct;
	if (enamet) delete enamet;
	if (eordt) delete eordt;
	pe_shared->v_exports=v;
	return g;
pe_read_error:
	errorbox("%s: PE export directory seems to be corrupted.", file->getname());
	if (efunct) delete efunct;
	if (enamet) delete enamet;
	if (eordt) delete eordt;
	delete g;
	delete m;
	delete v;
	return 0;
}

viewer_int htpeexports_int = {
	htpeexports_detect,
	htpeexports_init
};

/*
 *	CLASS ht_pe_export_viewer
 */

ht_pe_export_viewer::ht_pe_export_viewer(bounds *b, char *desc, htstream *file, int caps)
: ht_uviewer(b, desc, file, caps)
{
}

ht_pe_export_viewer::~ht_pe_export_viewer()
{
}

char *ht_pe_export_viewer::func(int i, int getname)
{
	switch (i) {
		case 2:
			if (getname) return "sortord"; else {
				((hts_pe_export_mask*)currentsub)->sortbyord();
				dirtyview();
			}
			return 0;
		case 3:
			if (getname) return "sortva"; else {
				((hts_pe_export_mask*)currentsub)->sortbyva();
				dirtyview();
			}
			return 0;
		case 4:
			if (getname) return "sortname"; else {
				((hts_pe_export_mask*)currentsub)->sortbyname();
				dirtyview();
			}
			return 0;
		case 5:
			if (getname) return "unsort"; else {
				((hts_pe_export_mask*)currentsub)->unsort();
				dirtyview();
			}
			return 0;
	}
	return ht_uviewer::func(i, getname);
}

void ht_pe_export_viewer::ref(dword id)
{
	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;
	int j=pe_shared->v_image->jump(id, 1, this);
	if (j) app->focus(pe_shared->v_image); else errorbox("can't follow: export VA %08x is not valid !", id);
}

/*
 *	CLASS hts_pe_export_mask
 */

hts_pe_export_mask::hts_pe_export_mask(htstream *file, int start)
: hts_mask(file, start)
{
	sort_ord=0;
	sort_va=0;
	sort_name=0;
	sort_path=0;
	sort=0;
}

hts_pe_export_mask::~hts_pe_export_mask()
{
	if (sort_ord) delete sort_ord;
	if (sort_va) delete sort_va;
	if (sort_name) delete sort_name;
}

int hts_pe_export_mask::getline(char *line, ADDR ofs, int yofs)
{
	if ((sort) && (sort_path) && (ofs>=PE_EXPORT_DESC_LINES)) {
		return hts_mask::getline(line, sort_path[ofs], yofs);
	} else {
		return hts_mask::getline(line, ofs, yofs);
	}
}

void hts_pe_export_mask::sortbyord()
{
	if (!sort_ord) {
		int c=masks.count();
		byvalue *so=(byvalue*)malloc(sizeof *so * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data*)masks.get(i);
			char *s=d->tagstring;
			dword v=hexw(s);
			so[i].key=v;
			so[i].value=i;
		}
		qsortbyvalue(so, PE_EXPORT_DESC_LINES, c-1);
		sort_ord=(dword*)malloc(sizeof *sort_ord * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			sort_ord[i]=so[i].value;
		}
		delete so;
	}
	sort=1;
	sort_path=sort_ord;
}

void hts_pe_export_mask::sortbyva()
{
	if (!sort_va) {
		int c=masks.count();
		byvalue *sv=(byvalue*)malloc(sizeof *sv * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data*)masks.get(i);
			char *s=d->tagstring;
			dword v=hexd(s+5);
			sv[i].key=v;
			sv[i].value=i;
		}
		qsortbyvalue(sv, PE_EXPORT_DESC_LINES, c-1);
		sort_va=(dword*)malloc(sizeof *sort_va * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			sort_va[i]=sv[i].value;
		}
		delete sv;
	}
	sort=1;
	sort_path=sort_va;
}

void hts_pe_export_mask::sortbyname()
{
	if (!sort_name) {
		int c=masks.count();
		byname *sn=(byname*)malloc(sizeof *sn * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data*)masks.get(i);
			char *s=d->tagstring;
			char *t=get_editable(s, 1, 0);
			int q=tag_get_vlen(t);
			t+=22;

			sn[i].key=t;
			sn[i].keylen=q;
			sn[i].value=i;
		}
		qsortbyname(sn, PE_EXPORT_DESC_LINES, c-1);
		sort_name=(dword*)malloc(sizeof *sort_name * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			sort_name[i]=sn[i].value;
		}
		delete sn;
	}
	sort=1;
	sort_path=sort_name;
}

void hts_pe_export_mask::unsort()
{
	sort=0;
}

/*
 *	class ht_pe_export_function
 */

ht_pe_export_function::ht_pe_export_function()
{
	name=0;
	ordinal=0;
	function_address=0;
}

ht_pe_export_function::ht_pe_export_function(int function_address, int ordinal)
{
	this->ordinal=ordinal;
	this->function_address=function_address;
	byname=0;
}

ht_pe_export_function::ht_pe_export_function(int function_address, int ordinal, char *name)
{
	this->ordinal=ordinal;
	this->name=strdup(name);
	this->function_address=function_address;
	byname=1;
}

ht_pe_export_function::~ht_pe_export_function()
{
	if ((byname) && (name)) delete name;
}
