/* 
 *	HT Editor
 *	htpehead.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htnewexe.h"
#include "htpe.h"
#include "htpehead.h"
#include "httag.h"
#include "htstring.h"
#include "viewers.h"

#include "pestruct.h"

#define EXPL_PE_MACHINES 			0x00001000
#define EXPL_PE_MACHINES_STR			 "00001000"
#define EXPL_PE_OPTIONAL_MAGICS 		0x00001001
#define EXPL_PE_OPTIONAL_MAGICS_STR 	 "00001001"
#define EXPL_PE_SUBSYSTEMS 			0x00001002
#define EXPL_PE_SUBSYSTEMS_STR 		 "00001002"

#define PE_MACHINE_UNKNOWN		   	0
#define PE_MACHINE_I386			   	0x14c   // Intel 386
#define PE_MACHINE_I486			   	0x14d   // Intel 486
#define PE_MACHINE_I586			   	0x14e   // Intel 586
#define PE_MACHINE_R3000		   		0x162   // MIPS little-endian, 0x160 big-endian
#define PE_MACHINE_R4000		   		0x166   // MIPS little-endian
#define PE_MACHINE_R10000		   	0x168   // MIPS little-endian
#define PE_MACHINE_ALPHA		   		0x184   // Alpha AXP
#define PE_MACHINE_SH3			   	0x1a2   // Hitachi SH3
#define PE_MACHINE_SH4		   	 	0x1a6   // Hitachi SH4
#define PE_MACHINE_ARM				0x1c0   // ARM
#define PE_MACHINE_POWERPC		   	0x1f0   // IBM PowerPC Little-Endian
#define PE_MACHINE_IA64				0x200   // Intel IA64
#define PE_MACHINE_MIPS16			0x266   // MIPS16
#define PE_MACHINE_68k				0x268   // Motorola 68k
#define PE_MACHINE_ALPHA_AXP_64		0x284   // Alpha AXP 64
#define PE_MACHINE_MIPSf				0x366   // MIPSf
#define PE_MACHINE_MIPS16f			0x466   // MIPS16f

int_hash pe_machines[] =
{
	{PE_MACHINE_UNKNOWN, "generic"},
	{PE_MACHINE_I386, "Intel 386"},
	{PE_MACHINE_I486, "Intel 486"},
	{PE_MACHINE_I586, "Intel 586"},
	{PE_MACHINE_R3000, "R3000"},
	{PE_MACHINE_R4000, "R4000"},
	{PE_MACHINE_R10000, "R10000"},
	{PE_MACHINE_ALPHA, "Alpha AXP"},
	{PE_MACHINE_SH3, "Hitachi SH3"},
	{PE_MACHINE_SH4, "Hitachi SH4"},
	{PE_MACHINE_ARM, "ARM"},
	{PE_MACHINE_POWERPC, "Power PC"},
	{PE_MACHINE_IA64, "Intel IA64"},
	{PE_MACHINE_MIPS16, "MIPS16"},
	{PE_MACHINE_68k, "Motorola 68k"},
	{PE_MACHINE_ALPHA_AXP_64, "Alpha AXP 64"},
	{PE_MACHINE_MIPSf, "MIPSf"},
	{PE_MACHINE_MIPS16f, "MIPS16f"},
	{0, 0}
};

int_hash pe_optional_magics[] =
{
	{0x107, "ROM image"},
	{0x10b, "PE/PE32"},
	{0x20b, "PE32+/PE64"},
	{0, 0}
};

int_hash pe_subsystems[] =
{
	{0, "generic"},
	{1, "native"},
	{2, "Windows GUI"},
	{3, "Windows CUI"},
	{7, "POSIX CUI"},
	{9, "Windows CE GUI"},
	{10, "EFI"},
	{11, "EFI/boot"},
	{12, "EFI/runtime"},
	{0, 0}
};

char *coffheader[]=
{
	"magic                                            "_DWORD("00000000"),
	"machine                                          "_WORD("00000004")" "HT_EXPL_CH EXPL_PE_MACHINES_STR,
	"number of sections                               "_WORD("00000006"),
	"time-date stamp                                  "_TIME("00000008"),
	"pointer to symbol table                          "_DWORD("0000000c"),
	"number of symbols                                "_DWORD("00000010"),
	"size of optional header                          "_WORD("00000014"),
	"characteristics                                  "_WORD("00000016"),
	0
};

char *pe32header[] = {
	"optional magic                                   "_WORD("00000018")" "HT_EXPL_CH EXPL_PE_OPTIONAL_MAGICS_STR,
	"major linker version                             "_BYTE("0000001a"),
	"minor linker version                             "_BYTE("0000001b"),
	"size of code                                     "_DWORD("0000001c"),
	"size of data                                     "_DWORD("00000020"),
	"size of bss                                      "_DWORD("00000024"),
	"entry point                                      "_DWORD("00000028"),
	"code base                                        "_DWORD("0000002c"),
	"data base                                        "_DWORD("00000030"),
	"image base                                       "_DWORD("00000034"),
	"section alignment                                "_DWORD("00000038"),
	"file alignment                                   "_DWORD("0000003c"),
	"major OS version                                 "_WORD("00000040"),
	"minor OS version                                 "_WORD("00000042"),
	"major image version                              "_WORD("00000044"),
	"minor image version                              "_WORD("00000046"),
	"major subsystem version                          "_WORD("00000048"),
	"minor subsystem version                          "_WORD("0000004a"),
	"Win32 version                                    "_DWORD("0000004c"),
	"size of image                                    "_DWORD("00000050"),
	"size of headers                                  "_DWORD("00000054"),
	"checksum                                         "_DWORD("00000058"),
	"subsystem                                        "_WORD("0000005c")" "HT_EXPL_CH EXPL_PE_SUBSYSTEMS_STR,
	"dll characteristics                              "_WORD("0000005e"),
	"stack reserve                                    "_DWORD("00000060"),
	"stack commit                                     "_DWORD("00000064"),
	"heap reserve                                     "_DWORD("00000068"),
	"heap commit                                      "_DWORD("0000006c"),
	"loader flags                                     "_DWORD("00000070"),
	"number of directory entries                      "_DWORD("00000074"),
	"--- directories ---",
	"export directory                             rva "_DWORD("00000078")", size "_DWORD("0000007c"),
	"import directory                             rva "_DWORD("00000080")", size "_DWORD("00000084"),
	"resource directory                           rva "_DWORD("00000088")", size "_DWORD("0000008c"),
	"exception directory                          rva "_DWORD("00000090")", size "_DWORD("00000094"),
	"security directory                           rva "_DWORD("00000098")", size "_DWORD("0000009c"),
	"base relocation table                        rva "_DWORD("000000a0")", size "_DWORD("000000a4"),
	"debug directory                              rva "_DWORD("000000a8")", size "_DWORD("000000ac"),
	"description string                           rva "_DWORD("000000b0")", size "_DWORD("000000b4"),
	"machine value (GP)                           rva "_DWORD("000000b8")", size "_DWORD("000000bc"),
	"thread local storage (TLS)                   rva "_DWORD("000000c0")", size "_DWORD("000000c4"),
	"load configuration directory                 rva "_DWORD("000000c8")", size "_DWORD("000000cc"),
	"bound import directory                       rva "_DWORD("000000d0")", size "_DWORD("000000d4"),
	"import address table (IAT)                   rva "_DWORD("000000d8")", size "_DWORD("000000dc"),
	"delay import descriptor                      rva "_DWORD("000000e0")", size "_DWORD("000000e4"),
	"COM+ runtime header                          rva "_DWORD("000000e8")", size "_DWORD("000000ec"),
	"reserved (15)                                rva "_DWORD("000000f0")", size "_DWORD("000000f4"),
	0
};

char *section[] = {
	"-------------------------------------------------------------------------",
	"name                                             "_CHAR("00000000") _CHAR("00000001") _CHAR("00000002") _CHAR("00000003") _CHAR("00000004") _CHAR("00000005") _CHAR("00000006") _CHAR("00000007"),
	"virtual size                                     "_DWORD("00000008"),
	"virtual address                                  "_DWORD("0000000c"),
	"size                                             "_DWORD("00000010"),
	"offset                                           "_DWORD("00000014"),
	"relocation table offset                          "_DWORD("00000018"),
	"line number table offset                         "_DWORD("0000001c"),
	"relocation count                                 "_WORD("00000020"),
	"line number count                                "_WORD("00000022"),
	"characteristics                                  "_DWORD("00000024"),
	0
};

int htpeheader_detect(htstream *file)
{
	return 1;
}

ht_view *htpeheader_init(bounds *b, htstream *file, ht_group *group)
{
	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;

	int h=get_newexe_header_ofs(file);
	ht_uviewer *v=new ht_uviewer(b, DESC_PE_HEADER, file, htvc_edit);
	hts_mask *m=new hts_mask(file, h);
	register_expl(EXPL_PE_MACHINES, pe_machines);
	register_expl(EXPL_PE_OPTIONAL_MAGICS, pe_optional_magics);
	register_expl(EXPL_PE_SUBSYSTEMS, pe_subsystems);
	char info[128];
	sprintf(info, "* PE header at offset %08lx", pe_shared->hdr_ofs);
	m->add_mask(info);
/* COFF header */
	m->add_mask("--- COFF header ---");
	m->add_mask_table(coffheader);
/* optional header */
	m->add_mask("--- optional header ---");
	word opt;
	file->seek(h+24);
	file->read(&opt, 2);
	switch (opt) {
		case 0x10b:
			pe_shared->is_pe32=1;
			m->add_mask_table(pe32header);
			break;
		default: {
			m->add_mask("optional magic                                   "_WORD("00000018")" "HT_EXPL_CH EXPL_PE_OPTIONAL_MAGICS_STR);
			m->add_mask("-------------------------------------------------------------------------");
			m->add_mask("Unsupported optional magic ! If you this message in an original");
			m->add_mask("(unmodified) file, please contact us (see help).");
		}
	}
/* section headers */
	int sc=0, os=0;
	file->seek(h+6);
	file->read(&sc, 2);
	file->seek(h+20);
	file->read(&os, 2);
	file->seek(h+os+24);
	for (int i=0; i<sc; i++) {
		int j=0;
		while (section[j]) {
			char *s=strdup(section[j]);
			char *e=s;
			while ((e=findtag(e))) {
				int f=tag_get_offset(e);
				f+=24+os+i*40;
				tag_set_offset(e, f);
				e++;
			}
			m->add_mask(s);
			delete s;
			j++;
		}
	}
	v->insertsub(m);
	return v;
}

viewer_int htpeheader_int = {
	htpeheader_detect,
	htpeheader_init
};
