/* 
 *	HT Editor
 *	htperes.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htdialog.h"
#include "htiobox.h"
#include "htnewexe.h"
#include "htobj.h"
#include "htpe.h"
#include "htperes.h"
#include "htstream.h"
#include "htstring.h"
#include "httree.h"
#include "pestruct.h"

int_hash restypes[] = {
	{0x0001, "cursors"},
	{0x0002, "bitmaps"},
	{0x0003, "icons"},
	{0x0004, "menus"},
	{0x0005, "dialogs"},
	{0x0006, "string tables"},
	{0x0007, "font directories"},
	{0x0008, "fonts"},
	{0x0009, "accelerators"},
	{0x000a, "custom resource data"},
	{0x000b, "message tables"},
	{0x000c, "group cursors"},
	{0x000e, "group icons"},
	{0x0010, "version information"},
	{0x0011, "dialog includes"},
	{0x0013, "pnp"},
	{0x0014, "vxd"},
	{0x0015, "animated cursors"},
	{ 0 }
};

int_hash languages[] = {
	{7,  "german"},
	{9,  "english"},
	{10, "spanish"},
	{12, "french"},
	{16, "italian"},
	{ 0 }
};

int htperesources_detect(htstream *file)
{
	int h=get_newexe_header_ofs(file);
	dword pe=0;
	file->seek(h);
	file->read(&pe, 2);
	if (pe==PE_MAGIC) {
		int e, f;
		file->seek(h+0x78+PE_DIRECTORY_ENTRY_RESOURCE*8);
		file->read(&e, 4);
		file->read(&f, 4);
		return e|f;
	}
	return 0;
}

/* declared outside of read_resource_dir() because of its recursion... */

htstream *peresource_file;
dword peresource_dir_ofs;
ht_static_treeview *peresource_tree;
char peresource_string[128];
pe_section_headers *peresource_section_headers;

void read_resource_dir(void *node, int ofs, int level)
{
	if (level>2) return; 		/* prevents from infinite recursion on currupt resource directories */
	PE_RESOURCE_DIRECTORY dir;
// get directory
	peresource_file->seek(peresource_dir_ofs+ofs);
	peresource_file->read(&dir, sizeof dir);
// get entries
	PE_RESOURCE_DIRECTORY_ENTRY entry;
	for (int i=0; i<dir.name_count+dir.id_count; i++) {
		peresource_file->seek(peresource_dir_ofs+ofs+sizeof dir+i*8);
		peresource_file->read(&entry, sizeof entry);
		char *rm=peresource_string;
		if (entry.offset_to_directory & 0x80000000) {
			int hasname=0;
			char *name;
			if (entry.name & 0x80000000) {
				hasname=1;
				peresource_file->seek(peresource_dir_ofs+entry.name & 0x7fffffff);
				name=getstrw(peresource_file);
			}
			PE_RESOURCE_DIRECTORY subdir;
			peresource_file->seek(peresource_dir_ofs+entry.offset_to_directory & 0x7fffffff);
			peresource_file->read(&subdir, sizeof subdir);
			if (hasname) {
				rm+=sprintf(rm, "%s [%d]", name, subdir.name_count+subdir.id_count);
			} else {
				char *s=(!level) ? s=matchhash((short)entry.name, restypes) : 0;
				if (s) {
					rm+=sprintf(rm, "ID %04x, %s [%d]", (short)entry.name, s, subdir.name_count+subdir.id_count);
				} else {
					rm+=sprintf(rm, "ID %04x [%d]", (short)entry.name, subdir.name_count+subdir.id_count);
				}
			}
			void *n=peresource_tree->add_child(node, peresource_string);
			read_resource_dir(n, entry.offset_to_directory & 0x7fffffff, level+1);
			if (hasname) delete name;
		} else {
			char *s=matchhash((char)entry.name, languages);
			if (s) {
				rm+=sprintf(rm, "resource, %s (%04x) ", s, (short)entry.name);
			} else {
				rm+=sprintf(rm, "resource, unknown language (%04x) ", (short)entry.name);
			}

			PE_RESOURCE_DATA_ENTRY data;
			peresource_file->seek(peresource_dir_ofs+entry.offset_to_directory);
			peresource_file->read(&data, sizeof data);

			dword dofs;
			if (pe_rva_to_ofs(peresource_section_headers, data.offset_to_data, &dofs)) {
				rm+=sprintf(rm, "offset %08lx", dofs);
			} else {
				rm+=sprintf(rm, "offset ? (rva %08lx, currupt)", data.offset_to_data);
			}
			sprintf(rm, " size %08lx", data.size);
			peresource_tree->add_child(node, peresource_string);
		}
	}
}

ht_view *htperesources_init(bounds *b, htstream *file, ht_group *group)
{
	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;

	ht_static_treeview *t = new ht_static_treeview(b, DESC_PE_RESOURCES);

	void *root;
/* get resource directory offset */
	/* 1. get resource directory rva */
	dword iofs;
	dword irva=pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_RESOURCE].address;
//	dword isize=pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_RESOURCE].size;
	/* 2. transform it into an offset */
	if (!pe_rva_to_ofs(&pe_shared->sections, irva, &iofs)) goto pe_read_error;

/*     char info[128];
	sprintf(info, "* PE resource directory at offset %08lx", iofs);
	m->add_mask(info);*/

	peresource_file=file;
	peresource_dir_ofs=iofs;
	peresource_tree=t;
	peresource_section_headers=&pe_shared->sections;

	root=t->add_child(0, "pe resources");

	read_resource_dir(root, 0, 0);

	t->adjust(root, true);
	t->update();
	t->setpalette(pal_grp->get(windialog_special_palette_str));
	return t;
pe_read_error:
	errorbox("%s: PE resource directory seems to be corrupted.", file->getname());
	delete t;
	return 0;
}

viewer_int htperesources_int = {
	htperesources_detect,
	htperesources_init
};
