/* 
 *	HT Editor
 *	htsearch.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htsearch.h"

#include <stdlib.h>
#include <string.h>

/*
 *	binary search algorithm
 */

#define SEARCH_BUF_SIZE 1024*256

int bufequal(byte *buf1, byte *buf2, int len)
{
	for (int i=0; i<len; i++) {
		if (buf1[i]!=buf2[i]) return 0;
	}
	return 1;
}

void bufdowncase(byte *buf, dword len)
{
	for (dword i=0; i<len; i++) {
		if ((buf[i]>='A') && (buf[i]<='Z')) buf[i]+=32;
	}
}

#include "htdebug.h"
#include "htiobox.h"

int search_str(htstream *file, dword start, dword end, dword pattern_len, byte *p, dword *result_ofs, ht_progress_indicator *progress_indicator, int caseinsensitive, int idle)
{
	int ggg=new_timer();
	start_timer(ggg);

	byte *pattern=(byte*)malloc(sizeof *pattern * pattern_len);
	memmove(pattern, p, sizeof *pattern * pattern_len);
	if (caseinsensitive) {
		bufdowncase(pattern, pattern_len);
	}
	byte *buf=(byte*)malloc(SEARCH_BUF_SIZE);
	byte *bptr=buf;
	dword b;
	dword o=start;
	int end_reached=0;
	dword next_progress_draw=0, next_progress_inc=(end-start)/100;
	if ((progress_indicator) && (!idle)) {
		progress_indicator->redraw();
		screen->show();
	}
	if (pattern_len>=4) {
/* optimized for patterns of length >= 4 (dword size) */
		dword pdword=*((dword*)pattern);
		byte pbyte=*pattern;
		file->seek(o);
		b=SEARCH_BUF_SIZE;
		if (b+o>end) b=end-o;
		b=file->read(buf, b);
		if (caseinsensitive) {
			bufdowncase(buf, b);
		}
		while (1) {
			bptr=buf;
			byte *ll=buf+b-pattern_len;
			while (bptr<=ll) {
				if (pbyte==*bptr) {
					if (pdword==*((dword*)bptr)) {
						if (bufequal(pattern+4, bptr+4, pattern_len-4)) {
							*result_ofs=o+bptr-buf;
							delete pattern;
							delete buf;
							return 1;
						}
					}
				}
				bptr++;
			}
			o+=b-pattern_len;
			if ((progress_indicator) && (o>=next_progress_draw) && (!idle)) {
				ht_progress_indicator_data d;
				d.indicator=(o-start)*100/(end-start);
				progress_indicator->setdata(&d);
				progress_indicator->redraw();
				screen->show();
				next_progress_draw+=next_progress_inc;
			}
			if (end_reached) break;
			file->seek(o);
			b=SEARCH_BUF_SIZE;
			if (b+o>end) {
				b=end-o;
				end_reached=1;
			}
			b=file->read(buf, b);
		}
	} else {
/* unoptimized */
		dword pbyte=*pattern;
		file->seek(o);
		b=SEARCH_BUF_SIZE;
		if (b+o>end) b=end-o;
		b=file->read(buf, b);
		if (caseinsensitive) {
			bufdowncase(buf, b);
		}
		while (1) {
			bptr=buf;
			while (bptr-buf+pattern_len<=b) {
				if (pbyte==*bptr) {
					if (bufequal(pattern, bptr, pattern_len)) {
						*result_ofs=o+bptr-buf;
						delete pattern;
						delete buf;
						return 1;
					}
				}
				bptr++;
			}
			if ((progress_indicator) && (o>=next_progress_draw) && (!idle)) {
				ht_progress_indicator_data d;
				d.indicator=(o-start)*100/(end-start);
				progress_indicator->setdata(&d);
				progress_indicator->redraw();
				screen->show();
				next_progress_draw+=next_progress_inc;
			}
			o+=b-pattern_len;
			if (end_reached) break;
			file->seek(o);
			b=SEARCH_BUF_SIZE;
			if (b+o>end) {
				b=end-o;
				end_reached=1;
			}
			b=file->read(buf, b);
		}
	}
	delete pattern;
	delete buf;

	stop_timer(ggg);
//	infobox("took %d msecs.", get_timer_msec(ggg));
	delete_timer(ggg);

	return 0;
}

/*
 *	CLASS ht_hexascii_search_mask
 */

ht_hexascii_search_mask::ht_hexascii_search_mask(bounds *b)
: ht_group(b, hto_selectable, "some hex/ascii search mask")
{
	bounds c;
/* ascii string */
	c.x=6;
	c.y=0;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield(&c, 64);
	insert(str);
/* ascii string label */
	c.x=0;
	c.w=5;
	c.h=1;
	ht_label *strlabel=new ht_label(&c, "~ascii", str);
	insert(strlabel);

/* hex string */
	c.x=6;
	c.y=2;
	c.w=40;
	c.h=1;
	ht_hexinputfield *hex=new ht_hexinputfield(&c, 64);
	hex->attach(str);
	insert(hex);
/* hex string label */
	c.x=0;
	c.w=5;
	c.h=1;
	ht_label *hexlabel=new ht_label(&c, "~hex", hex);
	insert(hexlabel);

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield(&c, 10);
//	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label(&c, "~from addr", range_start);
//	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield(&c, 10);
//	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label(&c, "~to addr", range_end);
//	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_collection *opts=new ht_string_collection();
	opts->insert("~case insensitive");
//	opts->insert("find all occurencies (~idle)");
	options=new ht_checkboxes(&c, opts);
	insert(options);
}

ht_hexascii_search_mask::~ht_hexascii_search_mask()
{
}

int ht_hexascii_search_mask::datasize()
{
	return sizeof (ht_hexascii_search_mask)+str->datasize()+
		range_start->datasize()+range_end->datasize();
}

void ht_hexascii_search_mask::getdata(void *buf)
{
	ht_hexascii_search_mask_data *d=(ht_hexascii_search_mask_data*)buf;
	byte *v=(byte*)&d->variable;
/* str data */
	str->getdata(v);
	d->str=((ht_strinputfield_data*)v)->pure.text;
	d->strlen=((ht_strinputfield_data*)v)->pure.textlen;
	v+=str->datasize();
/* range_start data */
	range_start->getdata(v);
	d->start=((ht_strinputfield_data*)v)->pure.text;
	d->startlen=((ht_strinputfield_data*)v)->pure.textlen;
	v+=range_start->datasize();
/* range_end data */
	range_end->getdata(v);
	d->end=((ht_strinputfield_data*)v)->pure.text;
	d->endlen=((ht_strinputfield_data*)v)->pure.textlen;
	v+=range_end->datasize();
/* options */
	ht_checkboxes_data d2;
	options->getdata(&d2);
	d->caseinsensitive=(d2.state>>0)&1;
	d->idle=(d2.state>>1)&1;
}

/*
 *	CLASS ht_casm_search_mask
 */

ht_casm_search_mask::ht_casm_search_mask(bounds *b)
: ht_group(b, hto_selectable, "some c/asm search mask")
{
	bounds c;
/* string */
	c.x=0;
	c.y=1;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield(&c, 64);
	insert(str);
/* string label */
	c.x=0;
	c.y=0;
	c.w=12;
	c.h=1;
	ht_label *strlabel=new ht_label(&c, "C/ASM ~string", str);
	insert(strlabel);
/* hint */
	c.x=0;
	c.y=2;
	c.w=b->w-2;
	c.h=1;
	insert(new ht_statictext(&c, "(example: \"hello\\n\\0\\077\\xd\" 'ho',011b,66o)", 0));

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield(&c, 10);
//	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label(&c, "~from addr", range_start);
//	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield(&c, 10);
//	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label(&c, "~to addr", range_end);
//	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_collection *opts=new ht_string_collection();
	opts->insert("~case insensitive");
//	opts->insert("find all occurencies (~idle)");
	options=new ht_checkboxes(&c, opts);
	insert(options);
}

ht_casm_search_mask::~ht_casm_search_mask()
{
}

int ht_casm_search_mask::datasize()
{
	return sizeof (ht_casm_search_mask)+str->datasize()+
		range_start->datasize()+range_end->datasize();
}

void ht_casm_search_mask::getdata(void *buf)
{
	ht_casm_search_mask_data *d=(ht_casm_search_mask_data*)buf;
	byte *v=(byte*)&d->variable;
/* str data */
	str->getdata(v);
	d->str=((ht_strinputfield_data*)v)->pure.text;
	d->strlen=((ht_strinputfield_data*)v)->pure.textlen;
	v+=str->datasize();
/* range_start data */
	range_start->getdata(v);
	d->start=((ht_strinputfield_data*)v)->pure.text;
	d->startlen=((ht_strinputfield_data*)v)->pure.textlen;
	v+=range_start->datasize();
/* range_end data */
	range_end->getdata(v);
	d->end=((ht_strinputfield_data*)v)->pure.text;
	d->endlen=((ht_strinputfield_data*)v)->pure.textlen;
	v+=range_end->datasize();
/* options */
	ht_checkboxes_data d2;
	options->getdata(&d2);
	d->caseinsensitive=(d2.state>>0)&1;
	d->idle=(d2.state>>1)&1;
/*     ht_strinputfield_data d;
	str->getdata(&d);
	((ht_casm_search_mask_data*)buf)->str=d.text;
	((ht_casm_search_mask_data*)buf)->strlen=d.textlen;
	range_start->getdata(&d);
	((ht_casm_search_mask_data*)buf)->start=d.text;
	((ht_casm_search_mask_data*)buf)->startlen=d.textlen;
	range_end->getdata(&d);
	((ht_casm_search_mask_data*)buf)->end=d.text;
	((ht_casm_search_mask_data*)buf)->endlen=d.textlen;
	ht_checkboxes_data d2;
	options->getdata(&d2);
	((ht_casm_search_mask_data*)buf)->caseinsensitive=(d2.state>>0)&1;
	((ht_casm_search_mask_data*)buf)->idle=(d2.state>>1)&1;*/
}

