/* 
 *	HT Editor
 *	htstream.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTSTREAM_H__
#define __HTSTREAM_H__

#include "global.h"

#include <stdio.h>

/* some stdio.h define getc/putc macros */
#undef getc
#undef putc

/*
 *	CLASS htstream
 */

class htstream {
protected:
	int error;
	void seterror(int error);
public:
	htstream();
	virtual ~htstream();
/* new */
	ABSTRACT virtual int read(void *buf, int size);
	ABSTRACT virtual int write(void *buf, int size);
	ABSTRACT virtual int seek(int pos);
	ABSTRACT virtual int tell();
	ABSTRACT virtual int size();
	ABSTRACT virtual char *getname();
	STUB virtual void flush();
	STUB virtual void invd();
	STUB virtual int dirty();
	STUB virtual int bytedirty(int ofs);
		int geterror();
};

/*
 *	CLASS htrawstream
 */

class htrawstream: public htstream {
protected:
	char *name, *mode;
	int isdirty;
public:
	FILE *file;

	htrawstream();
	htrawstream(const char *name, const char *mode);
	virtual ~htrawstream();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
	virtual int seek(int pos);
	virtual int tell();
	virtual int size();
	virtual char *getname();
	virtual int dirty();
};

/*
 *	CLASS htrwstream
 */

#include "htdata.h"

class htrwstream: public htrawstream {
protected:
	ht_tree *vmod;
	int frag_count;
/* new */
	void putc(unsigned char c);
	unsigned char getc();
	void init_buf();
	void done_buf();
public:
	htrwstream();
	htrwstream(const char *name);
	virtual ~htrwstream();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
	virtual int seek(int pos);
	virtual int tell();
	virtual void flush();
	virtual void invd();
	virtual int bytedirty(int ofs);
};

/*
 *	CLASS htmemstream
 */

class htmemstream: public htstream {
protected:
	int ofs;
	int pos;
	int bufsize;
	char *buf;
	int isdirty;

	void extendbuf();
public:
	htmemstream();
	htmemstream(int ofs, int size);
	virtual ~htmemstream();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
	virtual int seek(int pos);
	virtual int tell();
	virtual int size();
	virtual char *getname();
	virtual int dirty();
};

/*
 *	string stream functions
 */

char *getstrz(htstream *stream);
void putstrz(htstream *stream, char *str);

char *getstrl(htstream *stream);
void putstrl(htstream *stream, char *str);

char *getstrw(htstream *stream);
void putstrw(htstream *stream, char *str);

/*
 *   file system functions
 */

int canonicalize(char *in_file, char *out_file);

#endif /* !__HTSTREAM_H__ */
