/* 
 *	HT Editor
 *	htstring.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "global.h"
#include "htdebug.h"
#include "htstring.h"

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#ifndef HAVE_STRDUP

#include <string.h>

char *strdup(char *str)
{
	   if (str) {
			 int len=strlen(str)+1;
			 char *s=(char*)malloc(len);
			 memmove(s, str, len);
			 return s;
	   } else {
			 return 0;
	   }
}

#endif /* HAVE_STRDUP */

/*
--- C/ASM strings ---
	tokens:
		"..."		C-strings
		'...'		pascal strings
		#32			pascal chars
		32			decimal numbers
		0xff			hexadecimal numbers (C notation)
		077			octal numbers 		(C notation)
		ffh 			hexadecimal numbers (ASM notation)
		77o			octal numbers 		(ASM notation)
		10011110b		binary numbers 	(ASM notation)
	separators:
		commas and/or whitespaces
*/

#define MAX_CASM_STR_LEN	1024

byte casmstr[MAX_CASM_STR_LEN];

char _casmerrstr[128];

char *casmerrstr=_casmerrstr;

void whitespaces(char **str)
{
	while ((unsigned char)**str<=32) (*str)++;
}

int waitfor(char **str, char b)
{
	while (**str!=b) {
		if (!**str) return 0;
		(*str)++;
	}
	return 1;
}

int hexdigit(char a)
{
	if ((a>='0') && (a<='9')) {
		return a-'0';
	} else if ((a>='a') && (a<='f')) {
		return a-'a'+10;
	} else if ((a>='A') && (a<='F')) {
		return a-'A'+10;
	}
	return -1;
}

int bnstr2bin(char *str, char *p, int base, dword *v)
{
	*v=0;
	do {
		int c=hexdigit(*str);
		if ((c==-1) || (c>=base)) return 0;
		(*v)*=base;
		*v+=c;
		str++;
	} while (str<p);
	return 1;
}

void casmerrf(char *format, ...)
{
	va_list arg;
	va_start(arg, format);
	vsprintf(_casmerrstr, format, arg);
	va_end(arg);
}

byte *eval_casmstr(char *str, int *ret_strlen)
{
	char *origstr=str;
	byte *c=casmstr;
	whitespaces(&str);
	while (*str) {
		switch (*str) {
			case '"':
				str++;
				while (*str!='"') {
					if (!*str) {
						casmerrf("unterminated string (\"...\") at pos %d", str-origstr);
						return 0;
					} else if (*str=='\\') {
						str++;
						switch (*str) {
							case 'a': /* Alarm/bell, 7 */
								*(c++)='\a'; str++;
								break;
							case 'b': /* Backspace, 8 */
								*(c++)='\b'; str++;
								break;
							case 'e': /* Escape, 27 (1b) */
								*(c++)='\e'; str++;
								break;
							case 'f': /* form Feed, 12 (0c) */
								*(c++)='\f'; str++;
								break;
							case 'n': /* liNe feed, 10 (0a) */
								*(c++)='\n'; str++;
								break;
							case 'r': /* carriage Return, 13 (0d) */
								*(c++)='\r'; str++;
								break;
							case 't': /* (horizontal) Tab, 9 */
								*(c++)='\t'; str++;
								break;
							case 'v': /* Vertical tab, 11 */
								*(c++)='\v'; str++;
								break;
							case 'x': /* heX */ {
								dword v=0;
								str++;
								char *p=str+1;
								if (hexdigit(*str)==-1) {
									casmerrf("\\x used with no following hex digits at pos %d", str-origstr);
									return 0;
								}
								if (hexdigit(*(str+1))!=-1) p++;
								if (!bnstr2bin(str, p, 16, &v)) {
									casmerrf("syntax error at pos %d", str-origstr);
									return 0;
								}
								*(c++)=(byte)v;
								str=p;
								break;
							}
							default:
								if ((*str>='0') && (*str<='7')) {
									dword v=0, w;
									while (1) {
										if ((*str<'0') || (*str>'7')) break;
										w=v;
										v<<=3;
										v+=*str-'0';
										if (v>255) {
											v=w;
											break;
										}
										str++;
									}
									*(c++)=(byte)v;
								} else {
									*(c++)=*str;
								}
								break;
						}
					} else *(c++)=*(str++);
				}
				str++;
				break;
			case '\'': {
				char *s=++str;
				if (!waitfor(&str, '\'')) {
					casmerrf("unterminated string ('...')  at pos %d", str-origstr);
					return 0;
				}
				memmove(c, s, str-s);
				c+=str-s;
				str++;
				break;
			}
			case '#': {
				str++;
				char *p=str;
				while (!strchr("#, ", *p)) p++;
				dword v;
				if (!bnstr2bin(str, p, 10, &v)) {
					casmerrf("invalid string for character at pos %d", str-origstr);
					return 0;
				}
				if (v>255) {
					casmerrf("value out of range for character (#123) at pos %d", str-origstr);
					return 0;
				}
				*(c++)=(byte)v;
				str=p;
				break;
			}
			default: {
				int base=10;
				int t=0;
				char *p=str;
				while (!strchr("#, ", *p)) p++;
				if (strncmp("0x", str, 2)==0) {
					str+=2;
					base=16;
				} else {
					switch (*(p-1)) {
						case 'b':
							base=2;
							p--;
							t++;
							break;
						case 'o':
							base=8;
							p--;
							t++;
							break;
						case 'd':
							base=10;
							p--;
							t++;
							break;
						case 'h':
							base=16;
							p--;
							t++;
							break;
						default:
							if (*str=='0') base=8;
					}
				}
				dword v;
				if (!bnstr2bin(str, p, base, &v)) {
					casmerrf("invalid string for character at pos %d", str-origstr);
					return 0;
				}
				if (v>255) {
					casmerrf("value out of range (>255) for character at pos %d", str-origstr);
					return 0;
				}
				*(c++)=(byte)v;
				str=p+t;
				break;
			}
		}
		while (*str && (((unsigned char)*str<=32) || (*str==','))) str++;
	}
	if (ret_strlen) *ret_strlen=c-casmstr;
	*c=0;
	return casmstr;
}

dword hexb(char *s)
{
	char b=*s;
	b-='0';
	if (b>9) b-='a'-'0'-10;
	char c=*(s+1);
	c-='0';
	if (c>9) c-='a'-'0'-10;
	return (b<<4)+c;
}

dword hexw(char *s)
{
	return (hexb(s)<<8) | hexb(s+2);
}

dword hexd(char *s)
{
	return (hexw(s)<<16) | hexw(s+4);
}

int hex2int(char *s, dword *v)
{
	dword n=0;
	while (*s) {
		unsigned char b=*s;
		if ((b>='0') && (b<='9')) b-='0'; else
		if ((b>='a') && (b<='f')) b-='a'-10; else
		if ((b>='A') && (b<='F')) b-='A'-10; else return 0;
		n<<=4;
		n|=b;
		s++;
	}
	*v=n;
	return 1;
}

int str2int(char *s, dword *v)
{
}

char hexchars[17]="0123456789abcdef";

char *mkhexb(char *buf, byte d)
{
	*buf++=hexchars[(d>>4)&0xf];
	*buf++=hexchars[d&0xf];
	return buf;
}

char *mkhexw(char *buf, word d)
{
	*buf++=hexchars[(d>>12)&0xf];
	*buf++=hexchars[(d>>8)&0xf];
	*buf++=hexchars[(d>>4)&0xf];
	*buf++=hexchars[d&0xf];
	return buf;
}

char *mkhexd(char *buf, dword d)
{
	*buf++=hexchars[(d>>28)&0xf];
	*buf++=hexchars[(d>>24)&0xf];
	*buf++=hexchars[(d>>20)&0xf];
	*buf++=hexchars[(d>>16)&0xf];
	*buf++=hexchars[(d>>12)&0xf];
	*buf++=hexchars[(d>>8)&0xf];
	*buf++=hexchars[(d>>4)&0xf];
	*buf++=hexchars[d&0xf];
	return buf;
}

int eval_intexpr(char *str, dword *v, int (*plugin_proc)(char *s, dword *v))
{
}

