/* 
 *	HT Editor
 *	httag.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "global.h"
#include "htdebug.h"
#include "htstring.h"
#include "httag.h"

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

void tag_autotag(char *autotagstring, char *dest)
{
}

#define optimized_findtag

char *findtag(char *tagstring)
{
#ifdef optimized_findtag
	int *s=(int*)tagstring;
	int e;
	while (*s) {
		e=*(s++);
		for (int i=0; i<4; i++) {
			if ((e&0xff)=='\e') return (char*)tagstring;
			if ((e&0xff)==0) return 0;
			e=e>>8;
			tagstring++;
		}
	}
	return 0;
#else
	return strchr(tagstring, '\e');
#endif
}

int tag_get_len(char *tagstring)
{
	switch (tagstring[1]) {
		case HT_EDIT_BYTE:
			return HT_EDIT_BYTE_LEN;
		case HT_EDIT_WORD:
			return HT_EDIT_WORD_LEN;
		case HT_EDIT_DWORD:
			return HT_EDIT_DWORD_LEN;
		case HT_EDIT_TIME:
			return HT_EDIT_TIME_LEN;
		case HT_EDIT_CHAR:
			return HT_EDIT_CHAR_LEN;
		case HT_EXPL:
			return HT_EXPL_LEN;
		case HT_SEL: {
			int l=hexb(tagstring+10);
			return HT_SEL_LEN(l);
		}
//			return HT_SEL_LEN;
		case HT_GROUP:
			return HT_GROUP_LEN;
	}
	assert(0);
	return -1;
}

int tag_get_vlen(char *tagstring)
{
	switch (tagstring[1]) {
		case HT_EDIT_BYTE:
			return HT_EDIT_BYTE_VLEN;
		case HT_EDIT_WORD:
			return HT_EDIT_WORD_VLEN;
		case HT_EDIT_DWORD:
			return HT_EDIT_DWORD_VLEN;
		case HT_EDIT_TIME:
			return HT_EDIT_TIME_VLEN;
		case HT_EDIT_CHAR:
			return HT_EDIT_CHAR_VLEN;
		case HT_SEL: {
			int l=hexb(tagstring+10);
			return HT_SEL_VLEN(l);
		}
		default:
			return 0;
	}
}

int tag_get_slen(char *tagstring)
{
	switch (tagstring[1]) {
		case HT_EDIT_BYTE:
			return HT_EDIT_BYTE_VLEN;
		case HT_EDIT_WORD:
			return HT_EDIT_WORD_VLEN;
		case HT_EDIT_DWORD:
			return HT_EDIT_DWORD_VLEN;
		case HT_EDIT_TIME:
			return HT_EDIT_TIME_VLEN;
		case HT_EDIT_CHAR:
			return HT_EDIT_CHAR_VLEN;
		case HT_SEL: {
			return 1;
		}
		default:
			return 0;
	}
}

int tag_get_size(char *tagstring)
{
	assert(tagstring[0]=='\e');
	switch (tagstring[1]) {
		case HT_EDIT_BYTE:
			return HT_EDIT_BYTE_SIZE;
		case HT_EDIT_WORD:
			return HT_EDIT_WORD_SIZE;
		case HT_EDIT_DWORD:
			return HT_EDIT_DWORD_SIZE;
		case HT_EDIT_TIME:
			return HT_EDIT_TIME_SIZE;
		case HT_EDIT_CHAR:
			return HT_EDIT_CHAR_SIZE;
		case HT_SEL:
			return HT_SEL_SIZE;
		default:
			return 0;
	}
}

dword tag_get_offset(char *tagstring)
{
	assert(tagstring[0]=='\e');
	switch (tagstring[1]) {
		case HT_EDIT_BYTE:
		case HT_EDIT_WORD:
		case HT_EDIT_DWORD:
		case HT_EDIT_TIME:
		case HT_EDIT_CHAR:
		case HT_SEL:
			return hexd(tagstring+2);
	}
	assert(0);
	return 0;
}

int tag_get_value(char *tagstring)
{
	assert(tagstring[0]=='\e');
	switch (tagstring[1]) {
		case HT_EDIT_BYTE:
		case HT_EDIT_WORD:
		case HT_EDIT_DWORD:
		case HT_EDIT_TIME:
		case HT_EDIT_CHAR:
			return hexb(tagstring+10);
	}
	assert(0);
	return -1;
}

void tag_set_color(char *tagstring, int color)
{
	assert(tagstring[0]=='\e');
	switch (tagstring[1]) {
		case HT_EDIT_BYTE: {
			mkhexb(tagstring+12, color);
			break;
		}
		case HT_EDIT_WORD: {
			mkhexb(tagstring+14, color);
			break;
		}
		case HT_EDIT_DWORD: {
			mkhexb(tagstring+18, color);
			break;
		}
		case HT_EDIT_TIME: {
			mkhexb(tagstring+18, color);
			break;
		}
		case HT_EDIT_CHAR: {
			mkhexb(tagstring+12, color);
			break;
		}
		case HT_SEL: {
			mkhexb(tagstring+12, color);
			break;
		}
	}
}

void tag_set_offset(char *tagstring, dword offset)
{
	assert(tagstring[0]=='\e');
	switch (tagstring[1]) {
		case HT_EDIT_CHAR:
		case HT_EDIT_BYTE:
		case HT_EDIT_WORD:
		case HT_EDIT_DWORD:
		case HT_EDIT_TIME:
		case HT_SEL: {
			mkhexd(tagstring+2, offset);
			break;
		}
	}
}

void tag_set_value(char *tagstring, int value)
{
	assert(tagstring[0]=='\e');
	switch (tagstring[1]) {
		case HT_EDIT_BYTE: {
			mkhexb(tagstring+10, value);
			break;
		}
		case HT_EDIT_WORD: {
			mkhexw(tagstring+10, value);
			break;
		}
		case HT_EDIT_DWORD: {
			mkhexd(tagstring+10, value);
			break;
		}
		case HT_EDIT_TIME: {
			mkhexd(tagstring+10, value);
			break;
		}
		case HT_EDIT_CHAR: {
			mkhexb(tagstring+10, value);
			break;
		}
	}
}

int tag_is_editable(char *tagstring)
{
	assert(tagstring[0]=='\e');
	switch (tagstring[1]) {
		case HT_EDIT_CHAR:
		case HT_EDIT_BYTE:
		case HT_EDIT_WORD:
		case HT_EDIT_DWORD:
		case HT_EDIT_TIME:
			return 1;
		case HT_SEL:
			return 0;
	}
	assert(0);
	return -1;
}

int tag_getselid(char *tagstring)
{
	assert(tagstring[0]=='\e');
	if (tagstring[1]==HT_SEL) {
		return hexd(tagstring+2+8+2+2);
	} else return -1;
}

void tag_strcpy(char *dest, char *src)
{
    strcpy(dest, src);
}

char *tag_strdup(char *tagstring)
{
    return strdup(tagstring);
}

int tag_strlen(char *tagstring)
{
	int c=0;
	while (*tagstring) {
		if (tagstring[0]=='\e') {
			int s=tag_get_len(tagstring);
			c+=tag_get_vlen(tagstring);
			tagstring+=s;
		} else {
			c++;
			tagstring++;
		}
	}
	return c;
}

int count_editables_in_group(char *tagstring, int group)
{
	int c=0;
	tagstring=get_group(tagstring, group);
	while ((tagstring=findtag(tagstring))) {
		switch (tagstring[1]) {
			case HT_EDIT_BYTE:
				c++;
				tagstring+=HT_EDIT_BYTE_LEN;
				break;
			case HT_EDIT_WORD:
				c++;
				tagstring+=HT_EDIT_WORD_LEN;
				break;
			case HT_EDIT_DWORD:
				c++;
				tagstring+=HT_EDIT_DWORD_LEN;
				break;
			case HT_EDIT_TIME:
				c++;
				tagstring+=HT_EDIT_TIME_LEN;
				break;
			case HT_EDIT_CHAR:
				c++;
				tagstring+=HT_EDIT_CHAR_LEN;
				break;
			case HT_SEL:
				c++;
				tagstring+=HT_SEL_LEN(0);
				break;
			case HT_GROUP:
				return c;
			default:
				tagstring+=tag_get_len(tagstring);
				break;
		}
	}
	return c;
}

int count_editables(char *tagstring)
{
	int c=0;
	while ((tagstring=findtag(tagstring))) {
		switch (tagstring[1]) {
			case HT_EDIT_BYTE:
				c++;
				tagstring+=HT_EDIT_BYTE_LEN;
				break;
			case HT_EDIT_WORD:
				c++;
				tagstring+=HT_EDIT_WORD_LEN;
				break;
			case HT_EDIT_DWORD:
				c++;
				tagstring+=HT_EDIT_DWORD_LEN;
				break;
			case HT_EDIT_TIME:
				c++;
				tagstring+=HT_EDIT_TIME_LEN;
				break;
			case HT_EDIT_CHAR:
				c++;
				tagstring+=HT_EDIT_CHAR_LEN;
				break;
			case HT_SEL:
				c++;
				tagstring+=HT_SEL_LEN(0);
				break;
			default:
				tagstring+=tag_get_len(tagstring);
				break;
		}
	}
	return c;
}

int count_groups(char *tagstring)
{
	int c=1;
	while ((tagstring=findtag(tagstring))) {
		switch (tagstring[1]) {
			case HT_GROUP:
				c++;
				tagstring+=HT_GROUP_LEN;
				break;
			default:
				tagstring+=tag_get_len(tagstring);
				break;
		}
	}
	return c;
}

char *get_editable(char *tagstring, int n, int group)
{
	char *r=0;
	if (group) tagstring=get_group(tagstring, group);
	n++;
	while ((n) && (tagstring=findtag(tagstring))) {
		switch (tagstring[1]) {
			case HT_EDIT_BYTE:
				n--;
				r=tagstring;
				tagstring+=HT_EDIT_BYTE_LEN;
				break;
			case HT_EDIT_WORD:
				n--;
				r=tagstring;
				tagstring+=HT_EDIT_WORD_LEN;
				break;
			case HT_EDIT_DWORD:
				n--;
				r=tagstring;
				tagstring+=HT_EDIT_DWORD_LEN;
				break;
			case HT_EDIT_TIME:
				n--;
				r=tagstring;
				tagstring+=HT_EDIT_TIME_LEN;
				break;
			case HT_EDIT_CHAR:
				n--;
				r=tagstring;
				tagstring+=HT_EDIT_CHAR_LEN;
				break;
			case HT_SEL:
				n--;
				r=tagstring;
				tagstring+=HT_SEL_LEN(0);
				break;
			case HT_GROUP:
				return (char*)r;
			default:
				tagstring+=tag_get_len(tagstring);
				break;
		}
	}
	return (char*)r;
}

char *get_group(char *tagstring, int group)
{
	char *r=tagstring;
	while ((group) && (tagstring=findtag(tagstring))) {
		switch (tagstring[1]) {
			case HT_GROUP:
				group--;
				tagstring+=HT_GROUP_LEN;
				r=tagstring;
				break;
			default:
				tagstring+=tag_get_len(tagstring);
				break;
		}
	}
	return (char*)r;
}

/*void htstriptags(char *s)
{
	while ((s=findtag(s))) {
		int l=strlen(s);
		switch (s[1]) {
			case HT_EDIT_BYTE:
				memmove(s, s+10, l-10+1);
				memmove(s+2, s+4, l-2+1);
				break;
			case HT_EDIT_WORD:
				memmove(s, s+10, l-10+1);
				memmove(s+2, s+6, l-4+1);
				break;
			case HT_EDIT_DWORD:
				memmove(s, s+10, l-10+1);
				memmove(s+2, s+10, l-8+1);
				break;
			case HT_SEL:
				memmove(s, s+HT_SEL_LEN, l-HT_SEL_LEN+1);
				break;
			default:
				assert(0);
				break;
		}
	}
}*/

char *mktag_edit_byte(char *buf, ADDR ofs, byte d, byte color)
{
	*buf++='\e';
	*buf++=HT_EDIT_BYTE;
	buf=mkhexd(buf, ofs);
	buf=mkhexb(buf, d);
	buf=mkhexb(buf, color);
	return buf;
}

char *mktag_edit_word(char *buf, ADDR ofs, word d, byte color)
{
	*buf++='\e';
	*buf++=HT_EDIT_WORD;
	buf=mkhexd(buf, ofs);
	buf=mkhexw(buf, d);
	buf=mkhexb(buf, color);
	return buf;
}

char *mktag_edit_dword(char *buf, ADDR ofs, dword d, byte color)
{
	*buf++='\e';
	*buf++=HT_EDIT_DWORD;
	buf=mkhexd(buf, ofs);
	buf=mkhexd(buf, d);
	buf=mkhexb(buf, color);
	return buf;
}

char *mktag_edit_time(char *buf, ADDR ofs, dword d, byte color)
{
	*buf++='\e';
	*buf++=HT_EDIT_TIME;
	buf=mkhexd(buf, ofs);
	buf=mkhexd(buf, d);
	buf=mkhexb(buf, color);
	return buf;
}

char *mktag_edit_char(char *buf, ADDR ofs, byte d, byte color)
{
	*buf++='\e';
	*buf++=HT_EDIT_CHAR;
	buf=mkhexd(buf, ofs);
	buf=mkhexb(buf, d);
	buf=mkhexb(buf, color);
	return buf;
}

char *mktag_group(char *buf)
{
	*buf++='\e';
	*buf++=HT_GROUP;
	return buf;
}

char *mktag_sel(char *buf, dword unique_id, byte color, char *string)
{
	return mktag_ref(buf, unique_id, color, string, 0);
}

char *mktag_ref(char *buf, dword unique_id, byte color, char *string, dword id)
{
	int len=strlen(string);
	*buf++='\e';
	*buf++=HT_SEL;
	buf=mkhexd(buf, unique_id);
	buf=mkhexb(buf, len);
	buf=mkhexb(buf, color);
	buf=mkhexd(buf, id);
	strcpy(buf, string);
	buf+=len;
	return buf;
}
