/* 
 *	HT Editor
 *	htcurses.cc (LINUX implementation)
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include <unistd.h>

#include "global.h"
#include "htcurses.h"
#include "htdebug.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *	CLASS screendrawbuf
 */
 
drawbufch *buf;
WINDOW *win;
int use_colors, use_high_colors;
int is_xterm;
int cursorx, cursory; 

short colormap[64];

screendrawbuf::screendrawbuf()
{
	bounds b;
	buf=0;
	int colors[8]={ COLOR_BLACK, COLOR_BLUE, COLOR_GREEN, COLOR_CYAN, COLOR_RED, COLOR_MAGENTA, COLOR_YELLOW, COLOR_WHITE };

	cursorx=0;
	cursory=0;
	win=initscr();
	use_colors=0;
	use_high_colors=0;
	if (has_colors()) {
	    use_colors=1;
	    char *term=getenv("TERM");
	    int bold_support=0;
	    attr_t attrs;
	    short cur_color=1;
	    start_color();
/* FIXME: Does this work ???: test if the WA_BOLD attr can be set */
	    attr_on(WA_BOLD, 0);
	    attrs=WA_NORMAL;
	    attr_get(&attrs, &cur_color, 0);
	    bold_support=(attrs==WA_BOLD);
	    attr_off(WA_BOLD, 0);
	    
	    is_xterm=(strcmp(term, "linux") && strcmp(term, "console"));
	    if ( (!is_xterm) && (!bold_support)) {
		WARN("terminal is of type '%s' (non-x-terminal) but bold_test fails !", term);
	    }
	    if (bold_support) {
		    use_high_colors=1;
	    } else {
		    WARN("terminal only supports 8 foreground colors !");
	    }
	    for (int fg=0; fg<8; fg++) {
		    for (int bg=0; bg<8; bg++) {
			colormap[fg+bg*8]=fg+bg*8;
			init_pair(fg+bg*8, colors[fg], colors[bg]);
		    }
	    }
	    colormap[7]=0;
	    colormap[0]=7;
	    init_pair(7, COLOR_BLACK, COLOR_BLACK);
	} else {
	    WARN("terminal lacks color support !");
	}
	timeout(0);
	meta(win, 1);
	keypad(win, 1);
	nodelay(win, 1);
	noecho();
	ESCDELAY=0;
	
	b.x=0;
	b.y=0;
	b.w=getmaxx(win);
	b.h=getmaxy(win);
	setbounds(&b);

	show();
}

screendrawbuf::~screendrawbuf()
{
	endwin();
	if (buf) delete buf;
}

void screendrawbuf::printchar(int x, int y, int c, int ch)
{
	drawbufch *b=buf+x+y*size.w;
	b->ch=ch;
	b->c=c;
}

int screendrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int n=0;
	drawbufch *ch=buf+x+y*size.w;
	while ((*text) && (n<l)) {
		ch->ch=*text;
		ch->c=c;
		ch++;	
		text++;	
		n++;
	}
	return n;
}

int screendrawbuf::lprintw(int x, int y, int c, int l, int *text)
{
	int n=0;
	drawbufch *ch=buf+x+y*size.w;
	while ((*text) && (n<l)) {
		ch->ch=*text;
		ch->c=c;
		ch++;	
		text++;	
		n++;
	}
	return n;
}

void screendrawbuf::fill(int x, int y, int w, int h, int c, int chr)
{
	for (int i=0; i<h; i++) {
		drawbufch *ch=buf+x+(y+i)*size.w;
		for (int j=0; j<w; j++) {
			ch->ch=chr;
			ch->c=c;
			ch++;
		}
	}
}

void screendrawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf) delete buf;
	buf=(drawbufch *)malloc(sizeof *buf * size.w * size.h);
	fill(size.x, size.y, size.w, size.h, 7, ' ');
}

void screendrawbuf::drawbuffer(drawbuf *b, int x, int y)
{
	for (int iy=0; iy<b->size.h; iy++) {
		drawbufch *src=b->buf+iy*b->size.w;
		drawbufch *dest=buf+x+(y+iy)*size.w;
		memmove(dest, src, b->size.w * sizeof *dest);
	}
}

void screendrawbuf::show()
{
#if 1
	drawbufch *ch=buf;
	int c=-1;
	for (int iy=0; iy<size.h; iy++) {
			move(iy+size.y, size.x);
			for (int ix=0; ix<size.w; ix++) {
				if ((use_colors) && (ch->c!=c)) {
					c=ch->c;
					if (use_high_colors) {
						if (is_xterm && (c==8)) {
					/* some terminals can't display dark grey, so we take light grey instead... */
						    attrset(A_NORMAL);
						    c=7;
						} else {
						    if (c&8) attrset(A_BOLD); else attrset(A_NORMAL);
						}
					}
					color_set( colormap[(c&7) | ((c&(7<<4))>>1)], 0);
				}
				if ((((unsigned char)ch->ch>=0x20) && ((unsigned char)ch->ch<=0x7e)) || ((unsigned int)ch->ch>0xff)) addch(ch->ch); else addch(32);
				ch++;
			}
	}
#endif	
	curs_set(0);
	refresh();
	curs_set(1);
	move(cursory, cursorx);
//	setsyx(cursory, cursorx);
//	fprintf(stdout, "\033[%i%d;%dH", 0, cursory+1, cursorx+1);
//move(0,0);addch('5');
}

void screendrawbuf::getcursor(int *x, int *y)
{
	*x=cursorx;
	*y=cursory;
}

void screendrawbuf::hidecursor()
{
	setcursor(0, 0);
}

void screendrawbuf::rmove(int rx, int ry)
{
    /* screens are not movable */
}

void screendrawbuf::setcursor(int x, int y)
{
	cursorx=x;
	cursory=y;
}

void screendrawbuf::showcursor()
{
}

/*
 *	CLASS drawbuf
 */

drawbuf::drawbuf(bounds *b)
{
	buf=0;
	setbounds(b);
}

drawbuf::~drawbuf()
{
	if (buf) {
		delete buf;
	}
}

void drawbuf::printchar(int x, int y, int c, int ch)
{
	drawbufch *b=buf+(x-size.x)+(y-size.y)*size.w;
	b->ch=ch;
	b->c=c;
}

int drawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int ox=0;
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
	if (y<size.h) {
		while ((ox<size.w) && (*text) && (ox<l)) {
			b->ch=(byte)*text;
			b->c=c;
			text++;
			ox++;
			b++;
		}
	}
	return ox;
}

int drawbuf::lprintw(int x, int y, int c, int l, int *text)
{
	int ox=0;
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
	if (y<size.h) {
		while ((ox<size.w) && (*text) && (ox<l)) {
			b->ch=(byte)*text;
			b->c=c;
			text++;
			ox++;
			b++;
		}
	}
	return ox;
}

void drawbuf::fill(int x, int y, int w, int h, int c, int ch)
{
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
	for (int iy=y; iy<y+h; iy++) {
		for (int ix=x; ix<x+w; ix++) {
			b->ch=ch;
			b->c=c;
			b++;
		}
	}
}

void drawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf) delete buf;
	buf=(drawbufch*)malloc(sizeof *buf * size.w * size.h);
	fill(size.x, size.y, size.w, size.h, 7, ' ');
}

/*****************************************************************************
 ****************************************************************************/

genericdrawbuf::genericdrawbuf()
{
}

genericdrawbuf::~genericdrawbuf()
{
}

void genericdrawbuf::fill(int x, int y, int w, int h, int c, int ch)
{
}

int genericdrawbuf::printf(int x, int y, int c, char *format, ...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return print(x, y, c, buf);
}

int genericdrawbuf::print(int x, int y, int c, char *text)
{
	return lprint(x, y, c, 0x7fffffff, text);
}

int genericdrawbuf::printw(int x, int y, int c, int *text)
{
	return lprintw(x, y, c, 0x7fffffff, text);
}

void genericdrawbuf::printchar(int x, int y, int c, int ch)
{
}

int genericdrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	return 0;
}

int genericdrawbuf::lprintw(int x, int y, int c, int l, int *text)
{
	return 0;
}

void genericdrawbuf::rmove(int rx, int ry)
{
	size.x+=rx;
	size.y+=ry;
}

void genericdrawbuf::setbounds(bounds *b)
{
	size=*b;
}
