/* 
 *	HT Editor
 *	htcurses.h (LINUX implementation)
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTCURSES_H__
#define __HTCURSES_H__

#include "config.h"

#include CURSES_HDR
#undef clear
#undef move

#include "common.h"
#include <stdarg.h>

#define C_BLACK 0
#define C_BLUE 1
#define C_GREEN 2
#define C_CYAN 3
#define C_RED 4
#define C_MAGENTA 5
#define C_YELLOW 6
#define C_WHITE 7

#define C_LIGHT(c) (c+8)

#define C_FG(c) ((c)&15)
#define C_BG(c) (((c)>>4)&15)

#define COLOR(fg, bg) ((bg<<4)|fg)

#define CHAR_LINEV ACS_VLINE
#define CHAR_LINEH ACS_HLINE
#define CHAR_BORDERTL ACS_LTEE
#define CHAR_BORDERTR ACS_RTEE
#define CHAR_BORDERTU ACS_TTEE
#define CHAR_BORDERTD ACS_BTEE
#define CHAR_CORNERUL ACS_ULCORNER
#define CHAR_CORNERLL ACS_LLCORNER
#define CHAR_CORNERUR ACS_URCORNER
#define CHAR_CORNERLR ACS_LRCORNER
#define CHAR_FILLED_L ACS_CKBOARD	/* low */
#define CHAR_FILLED_M ACS_CKBOARD	/* medium */
#define CHAR_FILLED_H ACS_CKBOARD	/* high */
#define CHAR_FILLED_F ACS_BLOCK		/* full */
#define CHAR_FILLED_HU '#'
#define CHAR_FILLED_HL ' '
#define CHAR_QUAD_SMALL 'x'
#define CHAR_ARROW_UP ACS_UARROW
#define CHAR_ARROW_DOWN ACS_DARROW
#define CHAR_ARROWBIG_UP ACS_UARROW
#define CHAR_ARROWBIG_DOWN ACS_DARROW
#define CHAR_RADIO ACS_BULLET

struct drawbufch {
	int ch;
	int c;
};

/*
 *	CLASS genericdrawbuf
 */

class genericdrawbuf {
public:
	bounds size;

	genericdrawbuf();
	virtual ~genericdrawbuf();
	virtual void fill(int x, int y, int w, int h, int c, int ch);
		   int printf(int x, int y, int c, char *format, ...);
	virtual int print(int x, int y, int c, char *text);
	virtual int printw(int x, int y, int c, int *text);
	virtual void printchar(int x, int y, int c, int ch);
	virtual int lprint(int x, int y, int c, int l, char *text);
	virtual int lprintw(int x, int y, int c, int l, int *text);
	virtual void rmove(int rx, int ry);
	virtual void setbounds(bounds *b);
};

/*
 *	CLASS drawbuf
 */

class drawbuf: public genericdrawbuf {
public:
	drawbufch *buf;

	drawbuf(bounds *b);
	~drawbuf();
	virtual void fill(int x, int y, int w, int h, int c, int ch);
	virtual void printchar(int x, int y, int c, int ch);
	virtual int lprint(int x, int y, int c, int l, char *text);
	virtual int lprintw(int x, int y, int c, int l, int *text);
	virtual void setbounds(bounds *b);
};

class drawbuf;

class screendrawbuf: public genericdrawbuf {
public:
	screendrawbuf();
	~screendrawbuf();
	virtual void fill(int x, int y, int w, int h, int c, int ch);
	virtual void printchar(int x, int y, int c, int ch);
	virtual int lprint(int x, int y, int c, int l, char *text);
	virtual int lprintw(int x, int y, int c, int l, int *text);
	virtual void rmove(int rx, int ry);
	virtual void setbounds(bounds *b);
/* new */
	void drawbuffer(drawbuf *buf, int x, int y);
	void show();
	void getcursor(int *x, int *y);
	void hidecursor();
	void showcursor();
	void setcursor(int x, int y);
};

extern int xres, yres;

#endif /* !__HTCURSES_H__ */
